"""
vula constant values
"""
from typing import List

_LOG_FMT: str = "%(asctime)s: %(message)s"
_DATE_FMT: str = "%Y-%m-%d-%H:%M:%S"
_WG_INTERFACE: str = "vula"

_WG_SERVICES: List = [
    "vula-publish.service",
    "vula-organize.service",
    "vula-discover.service",
    "vula.slice",
]
_ORGANIZE_CACHE_BASEDIR: str = "/var/lib/vula-organize/"
_WG_PORT: int = 5354
_TABLE: int = 666
_FWMARK: int = 555
_IP_RULE_PRIORITY: int = 666
_SERVICE: str = "_opabinia._udp"
_DOMAIN: str = ".local."
_LABEL: str = "_opabinia._udp.local."
_DEFAULT_INTERFACE: str = _WG_INTERFACE
_B64ALPHABET: str = (
    "1234567890=/+abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
)
_ORGANIZE_CACHE_FILE: str = _ORGANIZE_CACHE_BASEDIR + "vula-organize-cache"
_ORGANIZE_CONF_FILE: str = _ORGANIZE_CACHE_BASEDIR + "vula-organize.yaml"
_ORGANIZE_KEYS_CONF_FILE: str = _ORGANIZE_CACHE_BASEDIR + "keys.yaml"
_ORGANIZE_HOSTS_FILE: str = _ORGANIZE_CACHE_BASEDIR + "hosts"
_ORGANIZE_UPDATE_TEMP: str = "vula-organize-peer-update-"
_DEFAULT_TABLE: int = 666

_ORGANIZE_DBUS_NAME: str = "local.vula.organize"
_DISCOVER_DBUS_NAME: str = "local.vula.discover"
_PUBLISH_DBUS_NAME: str = "local.vula.publish"
_DISCOVER_ALT_DBUS_NAME: str = "local.vula.discoveralt"
_PUBLISH_ALT_DBUS_NAME: str = "local.vula.publishalt"

# FIXME: these should not have the 1 in them, but pydbus publish() puts it
# there.  we should figure out how to make it work with the 1 removed from the
# object path (we want it in the interface name, but not the object path.)
_ORGANIZE_DBUS_PATH: str = "/local/vula/organize"
_DISCOVER_DBUS_PATH: str = "/local/vula/discover"
_PUBLISH_DBUS_PATH: str = "/local/vula/publish"
_DISCOVER_ALT_DBUS_PATH: str = "/local/vula/discoveralt"
_PUBLISH_ALT_DBUS_PATH: str = "/local/vula/publishalt"

_LINUX_MAIN_ROUTING_TABLE = 254

# IPv6 analysis: not ipv6 ready.
# please provide IPv6 Routes.
IPv4_GW_ROUTES = ('0.0.0.0/1', '128.0.0.0/1')

# Update interval for the tray in seconds
_TRAY_UPDATE_INTERVAL = 5



from __future__ import annotations

import copy
import traceback
from functools import reduce, wraps
from threading import Lock

from schema import Optional, Schema, Use

from .common import raw, schemattrdict, yamlfile, yamlrepr_hl


class Result(yamlrepr_hl, schemattrdict):
    """
    A result object contains the results of an event. It contains the event,
    and the actions, writes, triggers, and trigger_results which resulted from
    the event.

    Although log replay is not yet implemented, the event engine is designed
    such that replaying the events from a log of result objects should produce
    an identical state and an identical series of result objects (except for
    the trigger_results, which depend on the system's actual configuration
    state which exists outside of the state engine).
    """

    schema = Schema(
        {
            'event': Use(raw),
            'actions': Use(raw),
            'writes': Use(raw),
            Optional('triggers'): Use(raw),
            Optional('trigger_results'): Use(raw),
            Optional('error'): object,
            Optional('traceback'): str,
        },
    )

    default = dict(event=[], actions=[], writes=[], triggers=[])

    def __repr__(self):
        return "<Result\n  %s\n>" % (
            super(Result, self).__repr__().strip().replace("\n", "\n  "),
        )

    @property
    def ok(self):
        return not self.error

    @property
    def error(self):
        return self.get('error')

    @property
    def triggers(self):
        return self.setdefault('triggers', [])

    @property
    def trigger_results(self):
        return self.setdefault('trigger_results', [])

    @property
    def summary(self):
        if self.error:
            return "ERROR: %s" % (self.error,)
        else:
            return "OK: %s" % (" ".join(map(str, self.actions)))

    def add_triggers(self, **kw):
        for name, args in kw.items():
            self.triggers.append((name, args))

    def run_triggers(self, target):
        assert not self.trigger_results, "triggers should only be run once"
        for name, args in self.triggers:
            try:
                self.trigger_results.append(getattr(target, name)(*args))
            except Exception as ex:
                self.trigger_results.append(str(ex))
        return self


class Engine(schemattrdict, yamlfile):

    """
    This is a transactional state engine. Subclasses implement rules in the
    form of events and actions.

    An event engine using this class may be thought of as a pure function which
    takes a current state and a new event, and produces a new state.

    Subclasses should define @Engine.event methods and @Engine.action methods.
    Events should call one or more actions. Actions may mutate the state via
    write operations, and may call other actions, and may register triggers
    which will be run after the transaction is committed.

    The three built-in write methods are SET, ADD, and REMOVE. These methods,
    called from actions during an event transaction, are the only place where
    the engine state is allowed to be modified.

    If there are any exceptions during execution of the event and its actions,
    or if the state after all of the actions have been completed does not
    satisfy the schema, then none of the event's actions' write operations
    are applied.

    If the event does not have an error, after the new state is committed, the
    triggers are executed and their results are recorded in the event's
    result object. Triggers may modify state which exists outside of the state
    engine, and may also initiate new events.

    Calling an event will yield a Result object which contains a record of the
    event arguments and the resulting actions, writes, triggers, and trigger
    results, or contains the exception if one occurred.

    The state of the engine is only allowed to change through events which call
    actions which in turn call write methods. The new state must depend solely
    on the old state, and the event being processed. Therefore inputs from the
    outside world such as, for instance, the system time, need to be contained
    within an event to mutate the old state into the new state.  Similarly,
    side effects should only happen in triggers which are run after the
    successful processing of an event.
    """

    Result = Result

    def __init__(self, *a, **kw):
        self._lock = Lock()
        self.result = None
        self.next_state = None
        self.save = lambda *a: None
        self.debug_log = lambda *a: None
        self.trigger_target = None
        super(Engine, self).__init__(*a, **kw)

    def record(self, result):
        pass

    def event(method):
        """
        Decorator for event methods
        """

        assert method.__name__.startswith('event_')
        name = method.__name__.split('_', 1)[1]

        @wraps(method)
        def _method(self, *a, **kw):
            new_state = None
            res = self.Result(
                event=(name, *a),
                actions=[],
                writes=[],
                error=None,
            )
            error = None
            self._lock.acquire()
            try:
                self.next_state = copy.deepcopy(self._dict())
                self.result = res
                # run event method on a copy of our state
                method(self, *a, **kw)
                # confirm event produced a new valid state
                new_state = self.schema.validate(self.next_state)
                if raw(new_state) == raw(self):
                    self.debug_log("state unchanged")
                else:
                    # apply new state, cheating the ro_dict
                    dict.update(self, new_state)
                    self._as_dict = None  # part of careful ro_dict cheating
                    self.save()
            except Exception as ex:
                error = [ex, traceback.format_exc()]
                res = res._dict()
                res.update(error=error[0], traceback=error[1], triggers=[])
                res = self.Result(**res)
            finally:
                self.result = None
                self._lock.release()
            if self.trigger_target:
                res.run_triggers(self.trigger_target)
            self.record(res)
            self.debug_log(res)
            return res

        return _method

    def action(method):
        """
        Decorator for action methods
        """

        assert method.__name__.startswith('action_')
        name = method.__name__.split('_', 1)[1]

        @wraps(method)
        def _method(self, *a):
            self.result.actions.append((name,) + a)
            method(self, *a)

        return _method

    def write(method):
        """
        Decorator for write methods

        Writes are where the state gets changed. They should be called from
        action methods, which should be called from event methods.

        The data model here is a bit weird and still not stable, but roughly
        speaking there are three write methods and they operate on these
        types:

            set:
                - any type
            add:
                - lists (which it treats as sorted sets)
                - sets
                - dicts of whatever, when new value is a dict
                - dicts of bools, when new value is not a dict
            remove:
                - lists (which it treats as sorted sets)
                - sets
                - dicts of whatever
        """

        assert method.__name__.startswith('_')
        name = method.__name__[1:]

        @wraps(method)
        def _method(self, path, value):
            self.result.writes.append((name, path, value))

            if type(path) is str:
                path = path.split('.')
            if len(path) == 1:
                target = self.next_state
                key = path[0]
            else:
                target = reduce(lambda a, b: a[b], path[:-1], self.next_state)
                key = path[-1]

            method(self, target, key, raw(value))

        return _method

    @write
    def _SET(self, target, key, value):
        target[key] = value

    @write
    def _ADD(self, target, key, value):
        if isinstance(target[key], (list, tuple)):
            target[key] = type(target[key])(
                item for item in target[key] if raw(item) != value
            ) + type(target[key])((value,))
        elif isinstance(target[key], (set, frozenset)):
            target[key] = type(target[key])(
                frozenset(raw(target[key])) | set([value])
            )
        elif isinstance(target[key], dict):
            if isinstance(value, dict):
                target[key].update(value)
            else:
                target[key].update({value: True})
        else:
            raise ValueError("Can't add type: %r" % type(target[key]))

    @write
    def _REMOVE(self, target, key, value):
        if isinstance(target[key], (list, tuple)):
            target[key] = type(target[key])(
                item for item in raw(target[key]) if item != raw(value)
            )
        elif isinstance(target[key], (set, frozenset)):
            target[key] = type(target[key])(
                frozenset(raw(target[key])) - set([raw(value)])
            )
        elif isinstance(target[key], dict):
            del target[key][value]
        else:
            raise ValueError("Can't remove type: %r" % type(target[key]))


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import gettext
import json
import tkinter as tk

from vula.frontend import DataProvider
from vula.frontend.components import QRCodeLabel
from vula.frontend.constants import (
    BACKGROUND_COLOR,
    FONT,
    FONT_SIZE_HEADER,
    FONT_SIZE_TEXT_XXL,
    TEXT_COLOR_HEADER,
    TEXT_COLOR_WHITE,
)
from vula.peer import Descriptor

_ = gettext.gettext


class DescriptorOverlay(tk.Toplevel):
    def __init__(self, parent: tk.Tk) -> None:
        self.root = parent

    def openNewWindow(self) -> None:
        data = DataProvider()
        newWindow = tk.Toplevel(self.root)
        newWindow.wm_transient(self.root)

        # Root Window properties
        newWindow.title("Vula | Descriptor")
        newWindow.geometry("510x530")
        newWindow.configure(bg=BACKGROUND_COLOR)

        # Create the Frames for diffrent sections
        title_frame = tk.Frame(
            newWindow,
            bg=BACKGROUND_COLOR,
            width=510,
            height=70,
            pady=10,
            padx=10,
        )
        text_frame = tk.Frame(
            newWindow, bg=BACKGROUND_COLOR, width=510, height=50, padx=20
        )
        qr_frame = tk.Frame(
            newWindow, bg=BACKGROUND_COLOR, width=510, height=410
        )

        # Create the grid for the root TK
        newWindow.grid_rowconfigure(2, weight=1)
        newWindow.grid_columnconfigure(0, weight=1)

        # Place the Frames on the grid
        title_frame.grid(row=0, sticky="nw")
        text_frame.grid(row=1, sticky="nw")
        qr_frame.grid(row=2, sticky="nsew")

        # Add a Title to the Window
        tk.Label(
            title_frame,
            text="Descriptor",
            bg=BACKGROUND_COLOR,
            fg=TEXT_COLOR_HEADER,
            font=(FONT, FONT_SIZE_HEADER),
        ).pack()

        my_descriptors = {
            ip: Descriptor(d)
            for ip, d in json.loads(data.our_latest_descriptors()).items()
        }

        for ip, desc in my_descriptors.items():
            # IP Label
            label_ip = tk.Label(
                text_frame,
                text=ip,
                bg=BACKGROUND_COLOR,
                fg=TEXT_COLOR_WHITE,
                font=(FONT, FONT_SIZE_TEXT_XXL),
            )
            label_ip.pack()

            # Descriptor QR Code Image
            qr_data = "local.vula:desc:" + str(desc)
            qr_code = QRCodeLabel(parent=qr_frame, qr_data=qr_data, resize=2)
            qr_code.configure(background=BACKGROUND_COLOR)
            qr_code.pack()



import inspect
import os
import shutil
import sys
from functools import reduce, wraps

import click
import packaging.version as pkgv
from click.exceptions import Exit  # noqa: F401
from schema import Optional, Schema

"""
This file contains various click-related bits of vula. None of this is
strictly necessary; it is mostly used for debugging. It would be fine to remove
the cls=Debuggable from __main__.py and remove all of the DualUse
decorators from organize etc (using plain-old click.command to decorate the
organize class should work) and the software would still run the way we intend
it to be run in production.

So what do these magical click things do? They make it so that you can use -d
to do a post-mortem when there are exceptions (except maybe not in glib
threads... sad). They also make it so that you can run class methods as
subcommands, and so that there are subcommands for accessing attributes.

Eg, this:
    vula organize sync
will instantiate an organize object, loads its state, and then call the
object's sync() method. (Note that the way we actually intend sync to be called
is "vula sync" which calls Organize's sync method via dbus; this "vula
organize sync" method is instantiating an organize object from its state file
and calling the method on that.)

Or, this: vula -d peer.Descriptor --addrs 10.168.128.160 --c
vdDpRSGtsqvui8dox0iBq0SSp/zXSEU2dx5s5x+qcquSp0oIWgDuqJw50e9wrIuGub+SXzU0s5EIR
H49QmNYDw== --dt
86400 --e false --hostname wg-mdns-test3.local.  --pk
EqcQ5gYxzGtzg7B4xi83kLyfuSMp8Kv3cmAJMs12nDM= --port 5354 --r '' --s
T6htsKgwCp5MAXjPiWxtVkccg+K2CePsVa7uyUgxE2ouYKXg2qNL+0ut3sSbVTYjzFGZSCO6n80SR
aR+BIeOCg== --vf
1606276812 --vk 90Y5JGEjoklSDw51ffoHYXhWs49TTnCQ/D5yBbuf3Zg= valid

...will instantiate a Descriptor object and verify that its signature is
correct.

Note that the first example does not require -d, but the second one does:
automatic resolution of dotted attribute paths only happens in the top-level
command and only when debug mode is enabled. The first example, meanwhile, uses
the fact that the Organize class is a DualUse.

Another example which relies on Debuggable, and works with a function that
isn't decorated at all (types are inferred from type annotations):
    sudo vula -d configure._reconfigure_restart_systemd_services  --help

One more example, of chaining the attribute-getting functions:
    sudo vula organize state system_state current_subnets

Anyway, if this gets in the way we can get rid of some or all of it.
"""


class OrderedGroup(click.Group):
    def list_commands(self, ctx):
        return list(self.commands)


class Debuggable(OrderedGroup):

    """
    This is a subclass of click.Group which adds a --debug option which enables
    two features which are useful for debugging:

        - It will drop to a pdb.post_mortem shell after any unhandled exception

        - It allows for automatic commandline access to any function annotated
          with basic types (str, int, maybe others?)

    To use it, just decorate with @Debuggable.command() where you would
    otherwise be using @click.group()
    """

    def __init__(self, scope=None, **attrs):
        self.scope = scope or {}
        super(Debuggable, self).__init__(**attrs)
        self.params.append(
            click.Option(
                ('-d', '--debug/--no-debug'),
                show_default=True,
                hidden=not os.environ.get('DEBUG'),
                is_flag=True,
                default=(
                    os.environ.get('DEBUG')
                    and sys.stdin.isatty()
                    and sys.stdout.isatty()
                ),
                help="Drop to a pdb.post_mortem shell upon uncaught exception "
                "(default True if DEBUG env var is set and stdin/out are "
                "ttys, False otherwise)",
            )
        )

    def invoke(self, ctx):
        try:
            return super(Debuggable, self).invoke(ctx)
        except Exception as ex:
            if isinstance(ex, click.exceptions.ClickException):
                raise ex
            if isinstance(ex, click.exceptions.Exit):
                raise ex
            if ctx.params.get('debug'):
                import pdb
                import traceback

                tb = sys.exc_info()[2]
                traceback.print_tb(tb)
                print(
                    "stopping to allow inspecting exception:\n\n    "
                    "%r\n\ntype c to continue to "
                    "post-mortem frame, or q to quit.\n " % (ex,)
                )
                pdb.set_trace()
                print("pdb.post_mortem on %r:" % (ex,))
                pdb.post_mortem(tb)
            else:
                raise ex

    def get_command(self, ctx, command):

        if command in self.commands:
            return super(Debuggable, self).get_command(ctx, command)

        elif ctx.params.get('debug'):
            try:
                cmd = reduce(
                    lambda a, b: a.get(b)
                    if type(a) is dict
                    else getattr(a, b),
                    command.split('.'),
                    self.scope,
                )
            except Exception as ex:
                print(ex)
                return None
            if isinstance(cmd, click.core.Command):
                return cmd
            elif hasattr(cmd, 'cli') and isinstance(
                cmd.cli, click.core.Command
            ):
                return cmd.cli
            elif callable(cmd):
                return _click_command_from_annotated_function(cmd)


def _click_command_from_annotated_function(cmd):

    """
    This metaprogramming nonsense is only used for development, and hardly even
    that.

    There is actually a library called "autoclick" which presumably does a
    better job of doing what this function is doing.
    """

    @click.command()
    def wrapped(**kw):
        print(cmd(**kw))

    spec = inspect.getfullargspec(cmd)
    none = object()
    defaults = list(spec.defaults or ())
    defaults = [none] * (len(spec.args) - len(defaults)) + defaults
    assert len(defaults) == len(spec.args), "logic error"
    for name, default in zip(spec.args, defaults):
        if default is none:
            wrapped = click.argument(name, type=spec.annotations[name])(
                wrapped
            )
        else:
            wrapped = click.option(
                '--' + name,
                show_default=True,
                default=default,
                type=spec.annotations[name],
            )(wrapped)
    return wrapped


class DualUse(click.Group):

    """
    @DualUse.object() is a class decorator which enables class instances to be
    usable both as normal python objects and as click commandline functions.

    Methods which should be cli accessible should be decorated with
    @DualUse.method() or @property.
    """

    def __init__(self, *a, **kw):
        callback = kw.pop('callback')

        @wraps(callback)
        @click.pass_context
        def wrapper(ctx, *a, **kw):
            instance = callback(*a, **kw)
            if 'magic_instance' not in ctx.meta.setdefault(
                self.callback.__name__, {}
            ):
                ctx.meta[self.callback.__name__]['magic_instance'] = instance
            return instance

        super(DualUse, self).__init__(callback=wrapper, *a, **kw)

    @property
    def all_commands(self):
        """
        Return dictionary of DualUse methods and child classes, with
        self.commands applied on top of it.

        (FIXME: possibly wg.Interface is the only DualUse.object that actually
        uses self.commands/add_command? if so, this could be renamed
        'commands' if the link commands were ported to a nested dualuse
        object...)
        """
        res = {
            value.cli.name: value.cli
            for value in vars(self).values()
            if hasattr(value, 'cli') and value.cli is not self
        }
        res.update(**self.commands)
        return res

    def list_commands(self, ctx):
        return list(self.all_commands.keys()) + [
            name
            for name, value in vars(self.callback).items()
            if isinstance(value, property) and name != "__wrapped__"
        ]

    def get_command(self, ctx, name):
        """
        This got way out of hand. The 'else' branch in this method is just used
        for debugging, and not for everyday use. It allows accessing attributes
        recursively, so you can say things like "vula organize state
        system_state our_wg_pk" and it will print the pk. But, for certain
        objects, it hits max recursion depth, and I haven't figured out why
        yet. I reserve the right to remove this unsupported magic in the
        future.
        """
        res = self.all_commands.get(name)

        if res:
            return res

        else:

            @click.group(
                name=name,
                cls=type(self),
                invoke_without_command=True,
                help="Read %r property of %s object"
                % (name, self.callback.__name__),
            )
            @click.pass_context
            class _property_printer(object):
                @property
                def value(self_):
                    try:
                        return getattr(
                            ctx.meta[self.callback.__name__]['magic_instance'],
                            name,
                        )
                    except Exception as ex:
                        click.echo(ex)

                def __init__(self_, ctx):
                    if ctx.invoked_subcommand is None:
                        echo_maybepager(str(self_.value))

                def __getattr__(self_, name):
                    return getattr(self_.value, name)

            _property_printer.callback.__name__ += ':' + name

            return _property_printer

    @classmethod
    def method(cls, opts=(), *a, **kw):
        """
        Decorator to make methods of DualUse.object classes cli-accessible
        """

        def decorator(f):
            @wraps(f)
            def wrapper(*a, **kw):
                ctx = click.get_current_context()
                instance = ctx.meta[f.__qualname__.split('.')[0]][
                    'magic_instance'
                ]
                res = f(instance, *a, **kw)
                if res:
                    res = str(res)
                    if res[-1] == '\n':
                        res = res[:-1]
                    click.echo(res)

            wrapper.__doc__ = f.__doc__
            decos = opts + (click.command(*a, **kw),)
            wrapper = reduce(lambda a, b: b(a), decos, wrapper)
            f.cli = wrapper
            # note: returning undecorated function, which has click command
            # attached to it
            return f

        return decorator

    @classmethod
    def object(cls, *a, **kw):
        """
        Decorator which installs an object instantiation cli in the 'cli'
        attribute of a class.
        """

        def decorator(f):
            f.cli = wraps(f)(
                click.group(cls=cls, **kw)(schema2click_options(f))
            )
            return f

        return decorator


def _make_type(schema):
    class _type(click.ParamType):
        """
        Note that these click types have awful looking type names currently, as
        the name is literally the whole definition. This will hopefully look
        better when we upgrade to the new version of the schema module which
        allows schemas to have proper names, so we won't need to see the whole
        schema DSL source in our --help output.
        """

        name = str(schema)

        def convert(self, value, param, ctx):
            return Schema(schema).validate(value)

    return _type


def schema2click_options(f):
    if hasattr(f, 'schema'):
        for key, sub_schema in f.schema._schema.items():
            if type(key) is Optional:
                key = key._schema
            default = (getattr(f, 'default') or {}).get(key)
            _type = _make_type(sub_schema)
            f = click.option(
                "--%s" % (key,),
                type=_type(),
                default=default,
                show_default=True,
            )(f)
    return f


def red(s):
    """
    Formats the given string 's' to red foreground color.

    >>> red_string = "This text is red"
    >>> print(red(red_string))
    \x1b[31mThis text is red\x1b[0m
    """
    return click.style(s, fg="red")


def green(s):
    """
    Formats the given string 's' to green foreground color.

    >>> green_string = "This text is green"
    >>> print(green(green_string))
    \x1b[32mThis text is green\x1b[0m
    """
    return click.style(s, fg="green")


def blue(s):
    """
    Formats the given string 's' to blue foreground color.

    >>> blue_string = "This text is blue"
    >>> print(blue(blue_string))
    \x1b[34mThis text is blue\x1b[0m
    """
    return click.style(s, fg="blue")


def yellow(s):
    """
    Formats the given string 's' to yellow foreground color.

    >>> yellow_string = "This text is yellow"
    >>> print(yellow(yellow_string))
    \x1b[33mThis text is yellow\x1b[0m
    """
    return click.style(s, fg="yellow")


def bold(s):
    """
    Formats the given string 's' to be bold.

    >>> bold_string = "This text is bold"
    >>> print(bold(bold_string))
    \x1b[1mThis text is bold\x1b[0m
    """
    return click.style(s, bold=True)


def echo_maybepager(s):
    if s.count("\n") < shutil.get_terminal_size()[1]:
        click.echo(s)
    else:
        click.echo_via_pager(s)


def shell_complete_helper(fn):
    """
    This is a helper to maintain compatibility with both click 7.x and 8.x.

    We could pass the old "autocompletion" argument to click 7.x but instead we
    pass nothing because autocompletion didn't work there anyway.
    """
    if pkgv.parse(click.__version__) >= pkgv.parse('8.0.0'):
        return dict(shell_complete=fn)
    return {}



"""
vula frontend constant values
"""
WIDTH = 1050
HEIGHT = 800

BACKGROUND_COLOR = '#333941'
BACKGROUND_COLOR_CARD = '#1E272F'
BACKGROUND_COLOR_ENTRY = '#515456'
BACKGROUND_COLOR_ERROR = '#890000'
TEXT_COLOR_HEADER = '#BB89EE'
TEXT_COLOR_HEADER_2 = '#C0BABA'
TEXT_COLOR_WHITE = '#FFFFFF'
TEXT_COLOR_BLACK = '#000716'
TEXT_COLOR_GREEN = '#5BC960'
TEXT_COLOR_RED = '#EA1C1C'
TEXT_COLOR_GREY = '#838383'
TEXT_COLOR_PURPLE = '#BB89EE'
TEXT_COLOR_YELLOW = '#FFB800'
TEXT_COLOR_ORANGE = '#EF884F'

FONT = 'Roboto Regular'
FONT_SIZE_HEADER = 36 * -1
FONT_SIZE_HEADER_2 = 24 * -1
FONT_SIZE_TEXT_XS = 11 * -1
FONT_SIZE_TEXT_S = 12 * -1
FONT_SIZE_TEXT_M = 13 * -1
FONT_SIZE_TEXT_L = 14 * -1
FONT_SIZE_TEXT_XL = 16 * -1
FONT_SIZE_TEXT_XXL = 20 * -1

IMAGE_BASE_PATH = '/usr/share/icons/vula/'



import tkinter as tk

from vula.frontend.constants import (
    BACKGROUND_COLOR,
    BACKGROUND_COLOR_ERROR,
    FONT,
    FONT_SIZE_TEXT_XL,
    TEXT_COLOR_WHITE,
)


class PopupMessage:
    def showPopupMessage(title, text):
        popupRoot = tk.Toplevel()
        popupRoot.title(title)
        if title == "Error":
            backgroundcolor = BACKGROUND_COLOR_ERROR
        else:
            backgroundcolor = BACKGROUND_COLOR
        popupRoot.config(bg=backgroundcolor)
        popupRoot.after(2000, popupRoot.destroy)
        popupLabel = tk.Label(
            popupRoot,
            text=text,
            font=(FONT, FONT_SIZE_TEXT_XL),
            fg=TEXT_COLOR_WHITE,
            height=1,
            anchor="nw",
            bg=backgroundcolor,
            pady=10,
        )

        popupLabel.pack()
        popupRoot.geometry('400x50+400+525')
        popupRoot.mainloop()
        return



import tkinter as tk

from vula.frontend.constants import (
    BACKGROUND_COLOR,
    FONT,
    FONT_SIZE_TEXT_XL,
    FONT_SIZE_TEXT_XXL,
    TEXT_COLOR_WHITE,
)


class HelpOverlay(tk.Toplevel):
    def __init__(self, parent: tk.Tk) -> None:
        self.root = parent

    def openNewWindow(self) -> None:
        newWindow = tk.Toplevel(self.root)
        newWindow.wm_transient(self.root)

        # Root Window properties
        newWindow.title("Vula | Help")
        newWindow.geometry("530x530")
        newWindow.configure(bg=BACKGROUND_COLOR)

        # Create the Frames for diffrent sections
        text_frame = tk.Frame(
            newWindow, bg=BACKGROUND_COLOR, width=510, height=200, padx=20
        )

        # Create the grid for the root TK
        newWindow.grid_rowconfigure(0, weight=1)
        newWindow.grid_columnconfigure(0, weight=1)

        # Place the Frames on the grid
        text_frame.grid(row=0, sticky="nw")

        helpText = tk.Text(
            text_frame,
            bg=BACKGROUND_COLOR,
            fg=TEXT_COLOR_WHITE,
            font=(FONT, FONT_SIZE_TEXT_XL),
        )
        helpText.tag_configure("bold", font=(FONT, FONT_SIZE_TEXT_XXL, "bold"))
        helpText.insert("end", "Repair :\n", "bold")
        helpText.insert(
            "end", "Ensure that the system is configured correctly\n"
        )
        helpText.insert("end", "Rediscover :\n", "bold")
        helpText.insert(
            "end", "Tell organize to ask discover for more peers\n"
        )
        helpText.insert("end", "Release Gateway :\n", "bold")
        helpText.insert("end", "Stop using vula for the default route\n")
        helpText.pack()



"""
*vula* CTIDH interface functions.
"""
from hashlib import sha512
from typing import ByteString

from highctidh import ctidh  # noqa: F401
from hkdf import Hkdf

from .common import b64_bytes

# ctidh field size
ctidh_parameters = 512


def hkdf(raw_key: ByteString) -> str:
    """
    >>> s = hkdf(b"my_raw_key")
    >>> s
    'Y52eWgiYuPYtHlnqZpRqAG2USxILzRS57s61ePUdWO4='
    >>> type(s)
    <class 'str'>
    >>> hkdf("my_raw_key_string")
    Traceback (most recent call last):
     ...
    TypeError: ...
    """
    kdf = Hkdf(salt=None, input_key_material=raw_key, hash=sha512)
    key = kdf.expand(b"vula-organize-1", 32)
    psk = str(b64_bytes.with_len(32).validate(key))
    return psk


if __name__ == "__main__":
    import doctest

    doctest.testmod(verbose=1)



import csv
import fcntl
import hashlib
import json
import math
import random
import socket
import struct
import threading
import time
import zlib
from logging import Logger, getLogger
from queue import Empty, Queue
from typing import Optional

import click
import nacl.secret
import nacl.utils
import pydbus
from gi.repository import GLib

from .constants import (
    _ORGANIZE_DBUS_NAME,
    _ORGANIZE_DBUS_PATH,
    _PUBLISH_ALT_DBUS_NAME,
)


class Publish_Alt:

    dbus = '''
    <node>
      <interface name='local.vula.publishalt1.Control'>
        <method name='start'>
            <arg type='as' name='ip_addrs' direction='in'/>
        </method>
        <method name='stop'>
        </method>
      </interface>
    </node>
    '''

    def __init__(
        self,
        descriptor: str,
        verbose: bool,
        is_reply: bool,
        broadcast_mac: bytes,
        ethernet_type: bytes,
        hardware_type: bytes,
        protocol_type: bytes,
        hardware_size: bytes,
        protocol_size: bytes,
        zero_mac: bytes,
        interval_min: float,
        interval_max: float,
        op_code: bytes,
        arp_packet_max_length: int,
        formatting_string: str,
        hardware_address_ioctl_code: int,
        pa_address_code: int,
        time_interval: int,
    ):
        self.descriptor_cli = descriptor
        self.descriptor: dict = {}
        self.verbose = verbose
        self.is_reply = is_reply
        self.broadcast_mac = broadcast_mac
        self.ethernet_type = ethernet_type
        self.hardware_type = hardware_type
        self.protocol_type = protocol_type
        self.hardware_size = hardware_size
        self.protocol_size = protocol_size
        self.zero_mac = zero_mac
        self.interval_min = interval_min
        self.interval_max = interval_max
        self.op_code = op_code
        self.arp_packet_max_length = arp_packet_max_length
        self.formatting_string = formatting_string
        self.hardware_address_ioctl_code = hardware_address_ioctl_code
        self.pa_address_code = pa_address_code
        self.time_interval = time_interval
        self.log: Logger = getLogger()
        self.publish_entries: Queue = Queue()
        self.packets: Queue = Queue()
        self.active = False
        self.publishing_thread: threading.Thread
        self.packet_sending_thread: threading.Thread
        self.bus = pydbus.SystemBus()
        self.organize = self.bus.get(_ORGANIZE_DBUS_NAME, _ORGANIZE_DBUS_PATH)
        self.ip_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    def get_descriptors(self) -> None:
        """
        Update descriptors by querying organize
        """
        self.organize_descriptors = json.loads(
            self.organize.our_latest_descriptors()
        )

    def set_descriptors(self, sock: dict) -> None:
        """
        Set descriptor of ip
        """
        self.descriptor[sock["ip"]] = self.load_descriptor(sock["ip"])
        self.log.debug("Descriptors " + str(self.descriptor))

    def load_descriptor(self, ip: str) -> str:
        """
        Undo the JSON stuff done in organize and return descriptor
        """
        return " ".join(
            "%s=%s;" % kv
            for kv in sorted(self.organize_descriptors.get(ip).items())
        )

    def get_information(self, sock: dict) -> None:
        """
        Update specific information
        """
        sock["dest_ip"] = self.get_arp(sock)
        self.op_code = self.get_op_code()
        self.set_descriptors(sock)

    def run(self) -> None:
        """
        Main body for publishing thread
        """
        while self.active:
            self.get_descriptors()
            self.process_publish_entry(time.time())
            time.sleep(0.1)

    def process_publish_entry(self, t: float) -> bool:
        """
        Process queue of publishing entries. Go through FIFO
        queue and send descriptors. Reinsert into queue if periodic
        """
        try:
            publish_entry = self.publish_entries.get_nowait()
            if t - publish_entry["updated"] > self.time_interval:
                self.log.debug(publish_entry)
                # get dynamic information
                self.get_information(publish_entry)
                message = self.compress_and_encrypt(
                    self.descriptor[publish_entry["ip"]],
                    publish_entry["mac_addr"],
                )  # compress and encrypt message
                # generate packet header
                self.packets.put(
                    {
                        "packets": self.get_packets(publish_entry, message),
                        "socket": publish_entry["socket"],
                    }
                )
                publish_entry["updated"] = t
                if not publish_entry["periodic"]:
                    self.log.debug(
                        "dropping entry because it has been processed"
                    )
                else:
                    self.log.debug("adding entry back into queue")
                    self.publish_entries.put(publish_entry)
            else:
                self.publish_entries.put(publish_entry)
            return True
        except Empty:
            return False

    def get_packets(self, sock: dict, packet_payload: bytes) -> list:
        """
        Split message up into packets and return them for processing
        """
        self.log.debug("getting packets")
        packet_header = self.generate_header(sock)
        # segment encrypted payload into chunks of 18 bytes per packet
        # for a total packet length of 60 bytes
        packets = []
        number_of_packets = math.ceil(
            len(packet_payload)
            / (self.arp_packet_max_length - len(packet_header))
        )
        for i in range(number_of_packets):
            packet = (
                packet_header
                + packet_payload[
                    i
                    * ((self.arp_packet_max_length - len(packet_header))) : (
                        i + 1
                    )  # add corresponding packet chunk
                    * ((self.arp_packet_max_length - len(packet_header)))
                ]
            )
            packets.append(packet)
        return packets

    def send_packets(self) -> None:
        """
        Main body for sending thread. Sends packets in queue over the network.
        Wait for random duration of time between sending.
        """
        self.log.debug("setting up packet sending thread")
        while self.active:
            try:
                entry = self.packets.get_nowait()
                packets = entry["packets"]
                for i in range(len(packets)):
                    self.log.debug(f"Sending packet {i + 1} of {len(packets)}")
                    entry["socket"].send(packets[i])
                    time.sleep(
                        random.uniform(self.interval_min, self.interval_max)
                    )
            except Empty:
                pass
            time.sleep(0.5)

    def get_op_code(self) -> bytes:
        """
        Get operation code.
        """
        # set op code, 0x01 for request and 0x02 for reply
        self.log.debug(
            "Sending packet as reply."
            if self.is_reply
            else "Sending packet as request."
        )
        if self.is_reply:
            return b"\x00\x02"
        else:
            return b"\x00\x01"

    def get_ip(self, interface: str) -> str:
        """
        Get IP address of interface
        """
        return socket.inet_ntoa(
            fcntl.ioctl(
                self.ip_socket.fileno(),
                self.pa_address_code,
                struct.pack(
                    self.formatting_string,
                    interface[:15].encode("utf-8"),
                ),
            )[20:24]
        )

    def get_interface_name(self, ip: str) -> Optional[str]:
        """
        Return interface name associated withip
        """
        interfaces = socket.if_nameindex()
        for interface in interfaces:
            try:
                if ip == self.get_ip(interface[1]):
                    return interface[1]
            except OSError:
                pass
        return None

    def get_mac(self, sock: dict) -> bytes:
        """
        Get mac address of interface
        """
        return fcntl.ioctl(
            sock["socket"].fileno(),
            self.hardware_address_ioctl_code,
            struct.pack(
                self.formatting_string, bytes(sock["if_name"][:15], "utf-8")
            ),
        )[18:24]

    def compress_and_encrypt(self, msg: str, encryption_key: bytes) -> bytes:
        """
        Encrypt given message with key. Add timestamp for obfuscation
        """
        message = (
            str(int(time.time())) + msg
        )  # add unix timestamp to add some randomness to encrypted text
        hd = bytes(message, "utf-8")  # encode message into bytes
        hdc = zlib.compress(hd)  # compress with zlib
        key = hashlib.sha256(
            encryption_key
        ).digest()  # generate key from source mac
        box = nacl.secret.SecretBox(key)  # generate enryptor
        nonce = nacl.utils.random(
            nacl.secret.SecretBox.NONCE_SIZE
        )  # add default nonce
        self.log.debug("message length before compression: " + str(len(hd)))
        self.log.debug("message length after compression: " + str(len(hdc)))
        return box.encrypt(hdc, nonce)  # return encryped text

    def generate_header(self, sock: dict) -> bytes:
        """
        Assemble ARP header
        """
        packet_header = self.broadcast_mac
        packet_header += sock["mac_addr"]
        packet_header += self.ethernet_type
        packet_header += self.hardware_type
        packet_header += self.protocol_type
        packet_header += self.hardware_size
        packet_header += self.protocol_size
        packet_header += self.op_code
        packet_header += sock["mac_addr"]
        packet_header += socket.inet_aton(sock["ip"])
        packet_header += self.zero_mac
        packet_header += sock["dest_ip"]
        return packet_header

    def get_arp(self, sock: dict) -> bytes:
        """
        Pick random entry of ARP cache to be used as destination address.
        """
        with open("/proc/net/arp") as arp_table:  # read arp cache
            reader = list(
                csv.reader(arp_table, skipinitialspace=True, delimiter=" ")
            )
        dest_ip = b""
        arp_cache = [
            a[0] for a in reader[1:]
        ]  # skip header line and read ip field
        if len(arp_cache) == 0:
            dest_ip = (
                socket.inet_aton(self.get_ip(sock["if_name"]))[0:3] + b"\x01"
            )  # if chache is empty, take 0x01 address of own ip
        else:
            random_ip = random.choice(arp_cache)  # pick random ip
            dest_ip = socket.inet_aton(random_ip)
            self.log.debug("Setting random arp ip as destination.")
            self.log.debug("Arp cache: " + str(arp_cache))
            self.log.debug("Chosen Destination ip: " + str(random_ip))
        return dest_ip

    def start(self, ip_addrs: list[str]) -> None:
        """
        DBus exposed controlling function. Takes list of IPs and
        starts publishing process
        """
        self.log.info("Starting alternative publish")
        self.setup(ip_addrs)
        if not self.active:
            self.active = True
            self.packet_sending_thread = threading.Thread(
                target=self.send_packets
            )
            self.packet_sending_thread.start()
            self.publishing_thread = threading.Thread(target=self.run)
            self.publishing_thread.start()

    def stop(self) -> None:
        """
        DBus exposed controlling function. Stops discovering process
        """
        self.log.info("Stoping alternative publish")
        self.active = False
        self.clear_broadcast_queue()

    def setup(self, ip_addrs: list[str], periodically: bool = True) -> None:
        """
        Processes list of IP adresses and adds entries to be processed
        """
        self.log.debug("publish setup")
        for ip in ip_addrs:
            s = socket.socket(socket.AF_PACKET, socket.SOCK_RAW)
            if_name = self.get_interface_name(ip)
            s.bind((if_name, 0))
            socket_entry = {
                "socket": s,
                "periodic": periodically,
                "if_name": if_name,
                "ip": ip,
                "updated": 0,
            }
            socket_entry["mac_addr"] = self.get_mac(socket_entry)
            self.publish_entries.put(socket_entry)

    def clear_broadcast_queue(self) -> None:
        """
        Clears out all entries and closes the associated sockets
        """
        while not self.publish_entries.empty():
            socket_entry = self.publish_entries.get()
            socket_entry["socket"].close()
            del socket_entry["socket"]
        self.publish_entries = Queue()

    @classmethod
    def daemon(
        cls,
        descriptor,
        verbose,
        is_reply,
        broadcast_mac,
        ethernet_type,
        hardware_type,
        protocol_type,
        hardware_size,
        protocol_size,
        zero_mac,
        interval_min,
        interval_max,
        op_code,
        arp_packet_max_length,
        formatting_string,
        hardware_address_ioctl_code,
        pa_address_code,
        time_interval,
    ) -> None:
        """
        Sets up dbus
        """
        loop = GLib.MainLoop()

        publish = cls(
            descriptor,
            verbose,
            is_reply,
            broadcast_mac,
            ethernet_type,
            hardware_type,
            protocol_type,
            hardware_size,
            protocol_size,
            zero_mac,
            interval_min,
            interval_max,
            op_code,
            arp_packet_max_length,
            formatting_string,
            hardware_address_ioctl_code,
            pa_address_code,
            time_interval,
        )

        publish.log.info("dbus enabled")
        system_bus = pydbus.SystemBus()
        system_bus.publish(_PUBLISH_ALT_DBUS_NAME, publish)

        loop.run()


@click.command(short_help="Layer 2 alternate publish daemon")
@click.option(
    "--descriptor",
    required=False,
    type=str,
    help="descriptor to use for sending packets.",
)
@click.option(
    "--verbose",
    is_flag=True,
    default=False,
    help="show additional information.",
)
@click.option(
    "--is_reply",
    is_flag=True,
    default=False,
    help="Send reply packets instead of request.",
)
@click.option(
    "--broadcast_mac",
    default=b"\xff\xff\xff\xff\xff\xff",
    type=bytes,
    help="Broadcast mac address",
)
@click.option(
    "--ethernet_type",
    default=b"\x08\x06",
    type=bytes,
    help="Ethernet Type",
)
@click.option(
    "--hardware_type",
    default=b"\x00\x01",
    type=bytes,
    help="Hardware Type",
)
@click.option(
    "--protocol_type",
    default=b"\x08\x00",
    type=bytes,
    help="Protocol Type",
)
@click.option(
    "--hardware_size", default=b"\x06", type=bytes, help="Hardware Size"
)
@click.option(
    "--protocol_size", default=b"\x04", type=bytes, help="Protocol Size"
)
@click.option(
    "--zero_mac",
    default=b"\x00\x00\x00\x00\x00\x00",
    type=bytes,
    help="Empty mac address",
)
@click.option(
    "--interval_min",
    default=0.01,
    type=float,
    help="Minimum value of packet delay",
)
@click.option(
    "--interval_max",
    default=0.1,
    type=float,
    help="Maximum value of packet delay",
)
@click.option("--op_code", default=None, type=float, help="Operation Code")
@click.option(
    "--arp_packet_max_length",
    default=60,
    type=int,
    help="Max packet length",
)
@click.option(
    "--formatting_string",
    default="256s",
    type=str,
    help="Formatting string",
)
@click.option(
    "--hardware_address_ioctl_code",
    default=0x8927,
    type=int,
    help="IOCTL code to get hw address",
)
@click.option(
    "--pa_address_code",
    default=0x8915,
    type=int,
    help="IOCTL code to get PA address",
)
@click.option(
    "--time_interval", default=90, type=int, help="Time in between packets"
)
def main(
    descriptor,
    verbose,
    is_reply,
    broadcast_mac,
    ethernet_type,
    hardware_type,
    protocol_type,
    hardware_size,
    protocol_size,
    zero_mac,
    interval_min,
    interval_max,
    op_code,
    arp_packet_max_length,
    formatting_string,
    hardware_address_ioctl_code,
    pa_address_code,
    time_interval,
):
    arp_daemon = Publish_Alt.daemon(
        descriptor,
        verbose,
        is_reply,
        broadcast_mac,
        ethernet_type,
        hardware_type,
        protocol_type,
        hardware_size,
        protocol_size,
        zero_mac,
        interval_min,
        interval_max,
        op_code,
        arp_packet_max_length,
        formatting_string,
        hardware_address_ioctl_code,
        pa_address_code,
        time_interval,
    )
    arp_daemon.log.debug("main")


if __name__ == "__main__":
    main()



import unittest

import schema

from vula.common import raw
from vula.organize import OrganizeState, SystemState

from .test_peer import desc, mkk


class TestOrganizeEngine(unittest.TestCase):
    def setUp(self):
        self.maxDiff = 20000
        self.state = OrganizeState()
        # self.state.debug_log = lambda s: print(s)
        _ = self._assert_res_no_error(
            self.state.event_NEW_SYSTEM_STATE(
                SystemState(current_subnets={'10.0.0.0/24': ['10.0.0.9']})
            )
        )
        self._assert_res_no_error(
            self.state.event_USER_EDIT('SET', 'prefs.local_domains', ['local'])
        )

    def _assert_res_no_error(self, result: {}):
        if result.error:
            self.assertEqual(result.error, ())
        return result

    def _assert_res_actions(self, result, actions):
        if actions != [a[0] for a in result.actions]:
            raise Exception(
                "Unexpected result: \n\nExpected: %r\n\n"
                "Got: %r\n\nFull result:\n%r"
                % (
                    actions,
                    ["%s (%s)" % (a[0], a[1]) for a in result.actions],
                    result,
                )
            )
        return result

    def _proc_desc(self, actions=[], error=False, **kw):
        result = self.state.event_INCOMING_DESCRIPTOR(desc(**kw))
        if result.error and not error:
            self.assertEqual(result.error, ())
        if actions:
            self._assert_res_actions(result, actions)
        return result

    def _add_alice_ok(self):
        return self._proc_desc(
            actions=['ACCEPT_NEW_PEER'],
            hostname='alice.local',
            vk=mkk('alicevk'),
            pk=mkk('alicepk'),
            addrs='10.0.0.1',
        )

    def _add_bob_maybe(
        self, hostname='bob.local', addrs='10.0.0.2', pk=mkk('bobpk')
    ):
        return self._proc_desc(
            hostname=hostname, vk=mkk('bobvk'), pk=pk, addrs=addrs
        )

    def _add_alice_bob_same_ip(self):
        self._add_alice_ok()
        return self._add_bob_maybe(addrs='10.0.0.1')

    def _add_alice_bob_same_pk(self):
        self._add_alice_ok()
        # bob is using alice's pk
        return self._add_bob_maybe(pk=mkk('alicepk'))

    def _add_alice_bob_same_ip_and_pk(self):
        self._add_alice_ok()
        # bob is using alice's pk *and* ip. the audacity.
        return self._add_bob_maybe(pk=mkk('alicepk'), addrs='10.0.0.1')

    def _add_alice_bob_same_ip_and_hostname(self):
        self._add_alice_ok()
        # now bobvk is claiming alice's name and ip. this is like the
        # real-world scenario where a user has changed their vk.
        return self._add_bob_maybe(hostname='alice.local', addrs='10.0.0.1')

    def test_add_replace_unpinned_ip(self):
        self.assertEqual(self.state.prefs.pin_new_peers, False)
        self._assert_res_actions(
            self._add_alice_bob_same_ip(), ['REMOVE_PEER', 'ACCEPT_NEW_PEER']
        )
        self.assertEqual(
            self.state.peers.with_ip('10.0.0.1').name, 'bob.local'
        )

    def test_add_replace_unpinned_pk(self):
        self.assertEqual(self.state.prefs.pin_new_peers, False)
        self._assert_res_actions(
            self._add_alice_bob_same_pk(), ['REMOVE_PEER', 'ACCEPT_NEW_PEER']
        )

    def test_add_replace_unpinned_ip_and_pk(self):
        self.assertEqual(self.state.prefs.pin_new_peers, False)
        self._assert_res_actions(
            self._add_alice_bob_same_ip_and_pk(),
            ['REMOVE_PEER', 'ACCEPT_NEW_PEER'],
        )

    def test_add_replace_unpinned_ip_and_hostname(self):
        self.assertEqual(self.state.prefs.pin_new_peers, False)
        self._assert_res_actions(
            self._add_alice_bob_same_ip_and_hostname(),
            ['REMOVE_PEER', 'ACCEPT_NEW_PEER'],
        )

    def test_pin_protected_same_ip(self):
        self._assert_res_no_error(
            self.state.event_USER_EDIT('SET', 'prefs.pin_new_peers', True)
        )
        self.assertEqual(self.state.prefs.pin_new_peers, True)
        self._assert_res_actions(self._add_alice_bob_same_ip(), ['REJECT'])
        self.assertEqual(
            self.state.peers.with_ip('10.0.0.1').name, 'alice.local'
        )

    def test_pin_protected_same_pk(self):
        self._assert_res_no_error(
            self.state.event_USER_EDIT('SET', 'prefs.pin_new_peers', True)
        )
        self.assertEqual(self.state.prefs.pin_new_peers, True)
        self._assert_res_actions(self._add_alice_bob_same_pk(), ['REJECT'])

    def test_pin_protected_same_ip_and_pk(self):
        self._assert_res_no_error(
            self.state.event_USER_EDIT('SET', 'prefs.pin_new_peers', True)
        )
        self.assertEqual(self.state.prefs.pin_new_peers, True)
        self._assert_res_actions(
            self._add_alice_bob_same_ip_and_pk(), ['REJECT']
        )

    def test_pin_protected_same_ip_and_hostname(self):
        self._assert_res_no_error(
            self.state.event_USER_EDIT('SET', 'prefs.pin_new_peers', True)
        )
        self.assertEqual(self.state.prefs.pin_new_peers, True)
        self._assert_res_actions(
            self._add_alice_bob_same_ip_and_hostname(), ['REJECT']
        )

    def test_pin_protected_disabled(self):
        """
        Pin protection doesn't apply to disabled peers.

        TODO: test that they can't be reenabled without removing the conflict
        """
        self._assert_res_no_error(
            self.state.event_USER_EDIT('SET', 'prefs.pin_new_peers', True)
        )
        self.assertEqual(self.state.prefs.pin_new_peers, True)
        self._add_alice_ok()
        self._assert_res_no_error(
            self.state.event_USER_EDIT(
                'SET',
                [
                    'peers',
                    self.state.peers.with_hostname('alice.local').id,
                    'enabled',
                ],
                False,
            )
        )
        self._assert_res_actions(
            self._add_bob_maybe(addrs='10.0.0.1'), ['ACCEPT_NEW_PEER']
        )
        self.assertEqual(
            self.state.peers.with_ip('10.0.0.1').name, 'bob.local'
        )

    def test_bogon_announcement(self):
        self._proc_desc(
            hostname='alice.local',
            vk=mkk(1),
            addrs='10.0.0.1',
            actions=['ACCEPT_NEW_PEER'],
        )
        self._proc_desc(
            hostname='mallory.local',
            vk=mkk('3'),
            addrs='10.0.2.1',
            actions=['REJECT'],
        )
        self.assertEqual(len(self.state.peers), 1)

    def test_update(self):
        self.assertEqual(self.state.prefs.pin_new_peers, False)
        self._proc_desc(
            hostname='alice.local', vk=mkk(1), vf=1, addrs='10.0.0.1'
        )
        self._proc_desc(
            hostname='alice.local',
            vk=mkk(1),
            vf=2,
            addrs='10.0.0.1',
            actions=['UPDATE_PEER_DESCRIPTOR'],
        )
        self.assertEqual(self.state.peers[mkk(1)].descriptor.vf, 2)

    def test_replace_ip(self):
        "The one where Mallory takes the IP of unpinned peer Alice"
        self.assertEqual(self.state.prefs.pin_new_peers, False)
        self._proc_desc(
            hostname='alice.local',
            vk=mkk('alice'),
            pk=mkk('alicepk'),
            vf=1,
            addrs='10.0.0.1',
        )
        self._proc_desc(
            hostname='mallory.local',
            vk=mkk('mallory'),
            pk=mkk('mallorypk'),
            vf=1,
            addrs='10.0.0.1',
            actions=['REMOVE_PEER', 'ACCEPT_NEW_PEER'],
        )
        self.assertEqual(
            self.state.peers.with_ip('10.0.0.1').name, 'mallory.local'
        )
        with self.assertRaises(KeyError):
            self.state.peers.with_hostname('alice.local')

    def test_name_change_and_disable(self):
        s = self.state
        _ = self.state.event_USER_EDIT('SET', 'prefs.pin_new_peers', True)
        self.assertEqual(s.prefs.pin_new_peers, True)
        self._proc_desc(
            hostname='alice.local', vk=mkk('alice'), vf=1, addrs='10.0.0.1'
        )
        self._proc_desc(
            hostname='alice-1.local',
            vk=mkk('alice'),
            vf=2,
            addrs='10.0.0.1',
            actions=['UPDATE_PEER_DESCRIPTOR'],
        )
        self.assertEqual(
            s.peers.with_hostname('alice.local').name, 'alice-1.local'
        )
        self.assertEqual(
            s.peers.with_hostname('alice.local').nicknames,
            {'alice.local': True, 'alice-1.local': True},
        )
        self.assertEqual(
            s.peers.with_hostname('alice.local').enabled_names,
            ['alice-1.local', 'alice.local'],
        )
        self._assert_res_no_error(
            self.state.event_USER_EDIT(
                'SET',
                ['peers', mkk('alice'), 'nicknames', 'alice.local'],
                False,
            )
        )
        self.assertEqual(
            s.peers.with_ip('10.0.0.1').enabled_names,
            ['alice-1.local'],
        )

    def test_ignore_replay_unpinned(self):
        self._proc_desc(
            hostname='alice.local', vk=mkk(1), vf=2, addrs='10.0.0.2'
        )
        self._proc_desc(
            hostname='alice.local',
            vk=mkk(1),
            vf=1,
            addrs='10.0.0.1',
            actions=['IGNORE'],
        )
        self.assertEqual(
            raw(list(self.state.peers.with_hostname('alice.local').IPv4addrs)),
            ['10.0.0.2'],
        )

    def test_ignore_replay_pinned(self):
        _ = self.state.event_USER_EDIT('SET', 'prefs.pin_new_peers', True)
        self._proc_desc(
            hostname='alice.local', vk=mkk(1), vf=2, addrs='10.0.0.2'
        )
        self._proc_desc(
            hostname='alice.local',
            vk=mkk(1),
            vf=1,
            addrs='10.0.0.1',
            actions=['IGNORE'],
        )
        self.assertEqual(
            raw(list(self.state.peers.with_hostname('alice.local').IPv4addrs)),
            ['10.0.0.2'],
        )

    def test_state_validation(self):
        self._add_alice_ok()
        self._add_bob_maybe()
        # serialized dictionary of state:
        sd = self.state._dict()
        peers = self.state.peers  # shortcut
        # create bad state:
        sd['peers'][mkk('bobvk')]['petname'] = 'alice.local'
        with self.assertRaises(schema.SchemaError):
            # which means we're in an invalid state
            OrganizeState(sd)

        # make it valid again:
        sd['peers'][mkk('bobvk')]['petname'] = 'bob.local'
        OrganizeState(sd)
        # make it equal again:
        sd['peers'][mkk('bobvk')]['petname'] = ''
        self.assertEqual(
            OrganizeState(sd)._dict(),
            self.state._dict(),
        )
        # and then invalid due to an IP address conflict:
        sd['peers'][mkk("bobvk")]['IPv4addrs'].update(
            peers.with_hostname('alice.local')['IPv4addrs']
        )
        with self.assertRaises(schema.SchemaError):
            OrganizeState(sd)

        # make it valid again, by disabling a conflicting IP
        sd['peers'][peers.with_hostname('bob.local').id]['IPv4addrs'][
            list(peers.with_hostname('alice.local')['IPv4addrs'].keys())[0]
        ] = False
        OrganizeState(sd)

        # remove disabled IP to restore original state
        del sd['peers'][peers.with_hostname('bob.local').id]['IPv4addrs'][
            list(peers.with_hostname('alice.local')['IPv4addrs'].keys())[0]
        ]
        self.assertEqual(
            OrganizeState(sd)._dict(),
            self.state._dict(),
        )

        # create another invalid state,
        # where two peers are both set as the gateway
        sd['peers'][peers.with_hostname('alice.local').id][
            'use_as_gateway'
        ] = True
        sd['peers'][peers.with_hostname('bob.local').id][
            'use_as_gateway'
        ] = True
        with self.assertRaises(schema.SchemaError):
            OrganizeState(sd)

        # make it valid again
        sd['peers'][peers.with_hostname('bob.local').id][
            'use_as_gateway'
        ] = False
        OrganizeState(sd)

    def test_remove_nonlocal_unpinned(self):
        self._add_alice_ok()
        _ = self._assert_res_actions(
            self.state.event_NEW_SYSTEM_STATE(
                SystemState(self.state.system_state, current_subnets={})
            ),
            ['ADJUST_TO_NEW_SYSTEM_STATE', 'REMOVE_PEER'],
        )

    def test_user_edit_hostname_collision(self):
        self._add_alice_ok()
        self._assert_res_actions(
            self._add_bob_maybe(addrs='10.0.0.2'), ['ACCEPT_NEW_PEER']
        )
        res = self.state.event_USER_EDIT(
            'SET', ['peers', mkk('bobvk'), 'petname'], 'alice.local'
        )

        import packaging.version as pkgv
        import schema

        if pkgv.parse(schema.__version__) < pkgv.parse('0.7.3'):
            self.assertEqual(
                res.error.args[0], 'conflicting peers: {[peers].conflicts}'
            )
        else:
            self.assertEqual(
                res.error.args[0], 'conflicting peers: ' + mkk('alicevk')
            )


if __name__ == '__main__':
    unittest.main()



"""
Note: This module is currently unused.

>>> Network(essid={"a":True})
essid:
  a: true
<BLANKLINE>
>>> n = Network(bssid={"aa:bb:cc:dd:ee:ff":True})
>>> n
bssid:
  aa:bb:cc:dd:ee:ff: true
<BLANKLINE>
>>> bytes(list(n.bssid.keys())[0])
b'\xaa\xbb\xcc\xdd\xee\xff'
"""
from schema import Optional as Optional_
from schema import Schema

from .common import BSSID, ESSID, MACaddr, schemattrdict, yamlrepr


class Network(schemattrdict, yamlrepr):

    schema = Schema(
        {
            Optional_('essid'): {ESSID: bool},
            Optional_('gateway_MAC'): {MACaddr: bool},
            Optional_('bssid'): {BSSID: bool},
        }
    )



from unittest import TestCase, main


# fmt: off
class TestX25519(TestCase):

    # This fails on OpenBSD
    def test_x25519_keygen(self):
        import cryptography.hazmat.primitives.asymmetric.x25519
        from cryptography.hazmat.primitives.asymmetric.x25519 import (
            X25519PrivateKey,
        )

        keypair = X25519PrivateKey.generate()
        try:

            if (
                type(keypair)
                == cryptography.hazmat.backends.openssl.x25519._X25519PrivateKey  # noqa: E501
            ):
                self.assertIsInstance(
                    keypair,
                    cryptography.hazmat.backends.openssl.x25519._X25519PrivateKey,  # noqa: E501
                )
        except AttributeError:
            if (
                type(keypair)
                == cryptography.hazmat.bindings._rust.openssl.x25519.X25519PrivateKey  # noqa: E501
            ):
                self.assertIsInstance(
                    keypair,
                    cryptography.hazmat.bindings._rust.openssl.x25519.X25519PrivateKey,  # noqa: E501
                )


# fmt: on

if __name__ == '__main__':
    main()



import gettext
import tkinter as tk
from tkinter import messagebox, ttk
from typing import Literal

from vula.frontend import DataProvider, PeerType
from vula.frontend.constants import (
    BACKGROUND_COLOR,
    BACKGROUND_COLOR_ENTRY,
    FONT,
    FONT_SIZE_HEADER,
    FONT_SIZE_TEXT_M,
    IMAGE_BASE_PATH,
    TEXT_COLOR_BLACK,
    TEXT_COLOR_HEADER,
    TEXT_COLOR_ORANGE,
    TEXT_COLOR_WHITE,
)

from .popupMessage import PopupMessage

_ = gettext.gettext


class PeerDetailsOverlay(tk.Toplevel):
    data = DataProvider()

    def __init__(self, parent: tk.Frame, peer: PeerType) -> None:
        tk.Toplevel.__init__(self, parent)
        self.app = parent
        self.peer = peer

        self.id = peer["id"]
        if peer["name"]:
            self.name = peer["name"]
        else:
            self.name = peer["other_names"]

        self.title("Vula | Peer Details: " + self.name)
        self.geometry("520x500")
        self.config(bg=BACKGROUND_COLOR)
        self.resizable(True, True)
        self.return_value: Literal[
            'delete', 'pin_and_verify', 'rename', 'additional_ip', None
        ] = None
        self.display_peer_details()

    def display_peer_details(self) -> None:

        self.top_frame = tk.Frame(
            self,
            bg=BACKGROUND_COLOR,
            highlightthickness=0,
            width=500,
            padx=2,
            pady=2,
        )

        self.top_canvas = tk.Canvas(
            self.top_frame,
            bg=BACKGROUND_COLOR,
            width=500,
            bd=0,
            highlightthickness=0,
            relief="ridge",
        )

        self.yscrollbar = ttk.Scrollbar(
            self.top_frame,
            orient="vertical",
            command=self.top_canvas.yview,
        )

        frame = tk.Frame(
            self.top_canvas,
            bg=BACKGROUND_COLOR,
            width=600,
            height=870,
            highlightthickness=0,
        )
        frame.grid(row=0, sticky="ew")
        self.canvas = tk.Canvas(
            frame,
            bg=BACKGROUND_COLOR,
            height=870,
            width=600,
            bd=0,
            highlightthickness=0,
            relief="ridge",
        )
        self.canvas.place(x=0, y=0)

        # Packing and configuring
        self.top_canvas.pack(side="left", fill="y", expand="yes", anchor="nw")

        self.yscrollbar.pack(side="right", fill="y")

        self.top_canvas.configure(yscrollcommand=self.yscrollbar.set)
        self.top_canvas.bind(
            '<Configure>',
            lambda e: self.top_canvas.configure(
                scrollregion=self.top_canvas.bbox('all')
            ),
        )

        self.top_canvas.create_window((0, 0), window=frame, anchor="nw")

        self.top_frame.pack(
            fill="both", expand="yes", padx=(0, 0), pady=(0, 0), side="left"
        )

        self.top_frame.columnconfigure(0, weight=1)

        self.label = tk.Label(frame, text="test")

        # Title
        self.canvas.create_text(
            33.0,
            26.0,
            anchor="nw",
            text="Peer",
            fill=TEXT_COLOR_HEADER,
            font=(FONT, FONT_SIZE_HEADER),
        )

        # Delete
        self.button_delete_image = tk.PhotoImage(
            file=IMAGE_BASE_PATH + 'delete.png'
        )
        button_delete = tk.Button(
            frame,
            image=self.button_delete_image,
            borderwidth=0,
            highlightthickness=0,
            command=lambda: self.delete_peer(self.peer["id"], self.name),
            relief="sunken",
            background=BACKGROUND_COLOR,
            activebackground=BACKGROUND_COLOR,
            activeforeground=BACKGROUND_COLOR,
        )
        button_delete.place(x=450.0, y=34.0, width=39.0, height=31.0)

        # Name entry
        self.canvas.create_text(
            33.0,
            93.0,
            anchor="nw",
            text="name:",
            fill=TEXT_COLOR_WHITE,
            font=(FONT, FONT_SIZE_TEXT_M),
        )

        self.entry_image_name = tk.PhotoImage(
            file=IMAGE_BASE_PATH + 'edit_name_entry.png'
        )
        self.entry_image_name_bg = self.canvas.create_image(
            110.0, 126.5, image=self.entry_image_name
        )
        self.entry_name = tk.Entry(
            frame,
            bd=0,
            bg=BACKGROUND_COLOR_ENTRY,
            fg=TEXT_COLOR_BLACK,
            highlightthickness=0,
        )
        self.entry_name.insert(0, self.peer["name"])
        self.entry_name.place(x=45.5, y=114.0, width=129.0, height=23.0)

        # name save
        self.button_image_name = tk.PhotoImage(
            file=IMAGE_BASE_PATH + 'save.png'
        )
        button_name = tk.Button(
            frame,
            image=self.button_image_name,
            borderwidth=0,
            highlightthickness=0,
            command=lambda: self.edit_peer(),
            relief="sunken",
            background=BACKGROUND_COLOR,
            activebackground=BACKGROUND_COLOR,
            activeforeground=BACKGROUND_COLOR,
        )
        button_name.place(x=190.0, y=115.0, width=34.0, height=23.0)

        # Id
        self.canvas.create_text(
            33.0,
            168.0,
            anchor="nw",
            text="id:",
            fill=TEXT_COLOR_WHITE,
            font=(FONT, FONT_SIZE_TEXT_M),
        )

        self.canvas.create_text(
            33.0,
            194.0,
            anchor="nw",
            text=self.peer["id"],
            fill=TEXT_COLOR_ORANGE,
            font=(FONT, FONT_SIZE_TEXT_M),
        )

        self.button_image_copy_id = tk.PhotoImage(
            file=IMAGE_BASE_PATH + 'clipboard.png'
        )
        button_copy_id = tk.Button(
            frame,
            image=self.button_image_copy_id,
            borderwidth=0,
            highlightthickness=0,
            command=lambda: self.add_to_clipbaord(self.peer["id"]),
            relief="sunken",
            background=BACKGROUND_COLOR,
            activebackground=BACKGROUND_COLOR,
            activeforeground=BACKGROUND_COLOR,
        )
        button_copy_id.place(x=53.0, y=163.0, width=34.0, height=23.0)

        # Other names
        self.canvas.create_text(
            33.0,
            236.0,
            anchor="nw",
            text="other_names:",
            fill=TEXT_COLOR_WHITE,
            font=(FONT, FONT_SIZE_TEXT_M),
        )

        self.canvas.create_text(
            33.0,
            262.0,
            anchor="nw",
            text=self.peer["other_names"],
            fill=TEXT_COLOR_ORANGE,
            font=(FONT, FONT_SIZE_TEXT_M),
        )

        # Status
        self.canvas.create_text(
            33.0,
            304.0,
            anchor="nw",
            text="status:",
            fill=TEXT_COLOR_WHITE,
            font=(FONT, FONT_SIZE_TEXT_M),
        )

        self.canvas.create_text(
            33.0,
            330.0,
            anchor="nw",
            text=self.peer["status"],
            fill=TEXT_COLOR_ORANGE,
            font=(FONT, FONT_SIZE_TEXT_M),
        )

        self.button_image_pin_verify = tk.PhotoImage(
            file=IMAGE_BASE_PATH + 'pin_and_verify.png'
        )
        button_pin_verify = tk.Button(
            frame,
            image=self.button_image_pin_verify,
            borderwidth=0,
            highlightthickness=0,
            command=lambda: self.pin_verify(),
            relief="sunken",
            background=BACKGROUND_COLOR,
            activebackground=BACKGROUND_COLOR,
            activeforeground=BACKGROUND_COLOR,
        )
        button_pin_verify.place(x=78.0, y=299.0, width=92.0, height=23.0)

        # Endpoint
        self.canvas.create_text(
            33.0,
            372.0,
            anchor="nw",
            text="endpoint:",
            fill=TEXT_COLOR_WHITE,
            font=(FONT, FONT_SIZE_TEXT_M),
        )

        self.canvas.create_text(
            33.0,
            398.0,
            anchor="nw",
            text=self.peer["endpoint"],
            fill=TEXT_COLOR_ORANGE,
            font=(FONT, FONT_SIZE_TEXT_M),
        )

        # Allowed IP
        self.canvas.create_text(
            33.0,
            440.0,
            anchor="nw",
            text="allowed_ip:",
            fill=TEXT_COLOR_WHITE,
            font=(FONT, FONT_SIZE_TEXT_M),
        )

        self.canvas.create_text(
            33.0,
            466.0,
            anchor="nw",
            text=str(self.peer["allowed_ip"]),
            fill=TEXT_COLOR_ORANGE,
            font=(FONT, FONT_SIZE_TEXT_M),
        )

        # Latest signature
        self.canvas.create_text(
            33.0,
            508.0,
            anchor="nw",
            text="latest_signature:",
            fill=TEXT_COLOR_WHITE,
            font=(FONT, FONT_SIZE_TEXT_M),
        )

        self.canvas.create_text(
            33.0,
            534.0,
            anchor="nw",
            text=self.peer["latest_signature"],
            fill=TEXT_COLOR_ORANGE,
            font=(FONT, FONT_SIZE_TEXT_M),
        )

        # Latest handshake
        self.canvas.create_text(
            33.0,
            576.0,
            anchor="nw",
            text="latest_handshake:",
            fill=TEXT_COLOR_WHITE,
            font=(FONT, FONT_SIZE_TEXT_M),
        )

        self.canvas.create_text(
            33.0,
            602.0,
            anchor="nw",
            text=self.peer["latest_handshake"],
            fill=TEXT_COLOR_ORANGE,
            font=(FONT, FONT_SIZE_TEXT_M),
        )

        # Wg pubkey
        self.canvas.create_text(
            33.0,
            644.0,
            anchor="nw",
            text="wg_pubkey:",
            fill=TEXT_COLOR_WHITE,
            font=(FONT, FONT_SIZE_TEXT_M),
        )

        self.canvas.create_text(
            33.0,
            670.0,
            anchor="nw",
            text=self.peer["wg_pubkey"],
            fill=TEXT_COLOR_ORANGE,
            font=(FONT, FONT_SIZE_TEXT_M),
        )

        self.button_image_copy_wg_pubkey = tk.PhotoImage(
            file=IMAGE_BASE_PATH + 'clipboard.png'
        )
        button_copy_wg_pubkey = tk.Button(
            frame,
            image=self.button_image_copy_wg_pubkey,
            borderwidth=0,
            highlightthickness=0,
            command=lambda: self.add_to_clipbaord(self.peer["wg_pubkey"]),
            relief="sunken",
            background=BACKGROUND_COLOR,
            activebackground=BACKGROUND_COLOR,
            activeforeground=BACKGROUND_COLOR,
        )
        button_copy_wg_pubkey.place(x=108.0, y=639.0, width=34.0, height=23.0)

        # Allowed IPs
        self.canvas.create_text(
            33.0,
            712.0,
            anchor="nw",
            text="allowed_ips:",
            fill=TEXT_COLOR_WHITE,
            font=(FONT, FONT_SIZE_TEXT_M),
        )

        self.canvas.create_text(
            33.0,
            738.0,
            anchor="nw",
            text=self.peer["allowed_ips"],
            fill=TEXT_COLOR_ORANGE,
            font=(FONT, FONT_SIZE_TEXT_M),
        )

        # Name entry
        self.canvas.create_text(
            33.0,
            775.0,
            anchor="nw",
            text="Add additional IP to Peer:",
            fill=TEXT_COLOR_WHITE,
            font=(FONT, FONT_SIZE_TEXT_M),
        )

        self.entry_image_additional_ip = tk.PhotoImage(
            file=IMAGE_BASE_PATH + 'add_ip_to_peer_entry.png'
        )
        self.entry_image_additional_ip_bg = self.canvas.create_image(
            144.0, 808.5, image=self.entry_image_additional_ip
        )
        self.entry_additional_ip = tk.Entry(
            frame,
            bd=0,
            bg=BACKGROUND_COLOR_ENTRY,
            fg=TEXT_COLOR_BLACK,
            highlightthickness=0,
        )
        self.entry_additional_ip.place(
            x=45.5, y=796.0, width=197.0, height=23.0
        )

        # name save
        self.button_image_additional_ip = tk.PhotoImage(
            file=IMAGE_BASE_PATH + 'save.png'
        )
        button_additional_ip = tk.Button(
            frame,
            image=self.button_image_additional_ip,
            borderwidth=0,
            highlightthickness=0,
            command=lambda: self.add_additional_ip(),
            relief="sunken",
            background=BACKGROUND_COLOR,
            activebackground=BACKGROUND_COLOR,
            activeforeground=BACKGROUND_COLOR,
        )
        button_additional_ip.place(x=260.0, y=797.0, width=34.0, height=23.0)

    def delete_peer(self, peer_id: str, peer_name: str) -> None:
        # show message to ask user if they really
        # want to remove a peer

        box = messagebox.askyesno(
            _("Remove"),
            _("Remove this peer: ") + peer_name + " ?",
            parent=self,
        )

        if box:
            self.data.delete_peer(peer_id)
            self.return_value = "delete"
            self.destroy()

    def edit_peer(self) -> None:
        try:
            self.data.rename_peer(self.peer["id"], self.entry_name.get())
            self.return_value = "rename"
            self.destroy()
        except Exception:
            PopupMessage.showPopupMessage("Error", "Failed to edit peer")

    def pin_verify(self) -> None:
        # if name of peer was changed, we need other_names for pin_verify
        if self.peer["other_names"]:
            name = self.peer["other_names"]
        else:
            name = self.peer["name"]
        self.data.pin_and_verify(self.peer["id"], name)
        self.return_value = "pin_and_verify"
        self.destroy()

    def add_additional_ip(self) -> None:
        if len(self.entry_additional_ip.get()) == 0:
            return

        try:
            self.data.add_peer(self.peer["id"], self.entry_additional_ip.get())
            self.entry_additional_ip.delete(0, "end")
            self.return_value = "additional_ip"
            self.destroy()
        except Exception:
            PopupMessage.showPopupMessage(
                "Error", "Failed to add an additional ip address"
            )

    def add_to_clipbaord(self, text: str) -> None:
        self.clipboard_clear()
        self.clipboard_append(text)
        PopupMessage.showPopupMessage("Information", "Copied to clipboard")

    def show(
        self,
    ) -> Literal['delete', 'pin_and_verify', 'rename', 'additional_ip', None]:
        self.deiconify()
        self.wm_transient(self.app)
        self.wm_protocol("WM_DELETE_WINDOW", self.destroy)
        self.wait_window(self)
        return self.return_value



import threading
from ipaddress import ip_address, ip_network
from socket import AddressFamily

from pyroute2 import IPRoute, IPRSocket

from .constants import _LINUX_MAIN_ROUTING_TABLE, IPv4_GW_ROUTES
from .wg import Interface as WgInterface

# FIXME: find where the larger canonical version of this table lives
SCOPES = {0: 'global', 253: 'static'}


class Sys(object):
    """
    This  object provides all of the pyroute2-based system integration.
    Organize (and everything else) should only call pyroute2 via this object.

    This is currently our only implementation of our still-evolving Sys
    interface. We should reduce the number of public methods here to a minimum,
    and later we can reimplement this object using other means on other
    platforms.
    """

    def __init__(self, organize):
        self.organize = organize
        self.log = organize.log if organize else None
        self.wg_name = self.organize.interface if organize else None
        self.ipr = IPRoute()
        self.wgi = WgInterface(self.wg_name, ipr=self.ipr)
        self._monitor_thread = None
        self._stop_monitor = False

    def start_monitor(self):
        self._stop_monitor = False
        if self._monitor_thread is None:
            self._monitor_thread = threading.Thread(
                target=self._monitor
            )  # , args=(1,))
            self._monitor_thread.start()

    def get_stats(self):
        """
        Get the statistics

        >>> s = Sys(None)
        >>> type(s.get_stats())
        <class 'dict'>
        """
        self.wgi.query()
        stats = {peer.public_key: peer['stats'] for peer in self.wgi.peers}
        return stats

    def stop_monitor(self) -> None:
        """
        Stops the monitor
        """
        self._stop_monitor = True

    def _monitor(self):

        ip = IPRSocket()
        ip.bind()
        while True:
            msg = ip.get()
            if len(msg) != 1:
                self.log.info(
                    "BUG: got message with non-1 length %r which we didn't "
                    "expect ever happens",
                    msg,
                )
                continue
            event = msg[0].get('event')
            if event in [
                'RTM_DELADDR',
                'RTM_NEWADDR',
                'RTM_DELROUTE',
                'RTM_NEWROUTE',
            ]:
                self.log.debug("acting on netlink message: %r", msg)
                self.get_new_system_state()
            elif event == 'RTM_NEWNEIGH':
                # this happens often, so we don't even debug log it
                pass
            else:
                self.log.debug(
                    "ignoring netlink message type %r (%s bytes)",
                    event,
                    len(str(msg)),
                )
            if self._stop_monitor:
                self.log.info("Stopping netlink monitor thread")
                break
        self._monitor_thread = None
        ip.close()

    def _get_system_state(self):

        "WIP"

        addrs = self.ipr.get_addr()

        gateways = list(
            set(
                r.get_attrs('RTA_GATEWAY')[0]
                for r in self.ipr.get_routes()
                if r.get_attrs('RTA_GATEWAY')
            )
        )

        current_subnets = {}

        for a in addrs:
            addr = ip_address(dict(a['attrs'])['IFA_ADDRESS'])
            iface = dict(a['attrs']).get('IFA_LABEL', '')
            this_subnet = ip_network(
                "%s/%s" % (addr, a['prefixlen']), strict=False
            )
            if not any(
                [
                    iface.startswith(prefix)
                    for prefix in self.organize.prefs.iface_prefix_allowed
                ]
            ):
                continue

            if any(
                addr in subnet
                for subnet in self.organize.prefs.subnets_allowed
            ) and not any(
                addr in subnet
                for subnet in self.organize.prefs.subnets_forbidden
            ):
                current_subnets.setdefault(this_subnet, []).append(addr)

        return current_subnets, gateways

    def get_new_system_state(self):
        return self.organize.get_new_system_state()

    def sync_interface(self, dryrun=False):
        return self.wgi.sync_interface(
            private_key=str(self.organize._keys.wg_Curve25519_sec_key),
            listen_port=self.organize.port,
            fwmark=self.organize.fwmark,
            dryrun=dryrun,
        )

    def sync_peer(self, vk: str, dryrun: bool = False):
        """
        Syncs peer's wg config and routes. Returns a string.
        """
        # IPv6 analysis: not ipv6 ready.
        # Please enhance this function to support ipv6
        peer = self.organize.peers[vk]
        res = []
        if peer.enabled:
            self.log.debug("syncing enabled peer %s", peer.name)
            ctidh_psk = self.organize.ctidh_dh(peer.descriptor.c)
            res.append(
                self.wgi.apply_peerconfig(peer.wg_config(ctidh_psk), dryrun)
            )
            res.append(
                self.sync_routes(
                    peer.routes,
                    table=self.organize.table,
                    dryrun=dryrun,
                )
            )
            if peer.use_as_gateway:
                res.append(
                    self.sync_routes(
                        [*IPv4_GW_ROUTES],
                        table=_LINUX_MAIN_ROUTING_TABLE,
                        dryrun=dryrun,
                    )
                )
            self.log.debug("organize.Peer.sync result: %r", res)
        else:
            # FIXME: this should go away with triggers, but hasn't yet?
            self.remove_unknown()
        res = filter(None, res)
        return "\n".join(res)

    def sync_iprules(self, dryrun=False):
        routing_table = self.organize.table
        mark = self.organize.fwmark
        priority = self.organize.ip_rule_priority
        not_flag = 0x02
        res = []
        ip_version = {AddressFamily.AF_INET: "4", AddressFamily.AF_INET6: "6"}
        for family in ip_version.keys():
            existing_rule = [
                rule
                for rule in self.ipr.get_rules(family=family)
                if dict(rule['attrs'])['FRA_TABLE'] == routing_table
                and dict(rule['attrs'])['FRA_FWMARK'] == mark
                and dict(rule['attrs'])['FRA_PRIORITY'] == priority
                and rule['header']['flags'] == not_flag
            ]
            if existing_rule:
                self.log.debug(
                    "Found expected existing rule: (%s)", existing_rule
                )
            else:
                if not dryrun:
                    self.ipr.rule(
                        'add',
                        table=routing_table,
                        priority=priority,
                        fwmark=mark,
                        family=family,
                        flags=not_flag,
                    )
                res.append(
                    "ip -{af} rule add not from all fwmark 0x{mark:x} "
                    "lookup {table}".format(
                        af=ip_version[family], mark=mark, table=routing_table
                    )
                )
        return res

    def remove_wg_peer(self, pk, dryrun=False):
        return self.wgi.apply_peerconfig(
            dict(public_key=pk, remove=True), dryrun
        )

    def remove_routes(self, dests, table=None, dev=None, dryrun=False):
        """
        Idempotently remove route(s)

        Dests is a list of cidr notation strings.

        Returns a string describing what was (or would be) done.
        """
        if table is None:
            table = self.organize.table
        if dev is None:
            dev = self.wg_name
        res = []
        for route in self.get_route_entries(dests, table, dev):
            if not dryrun:
                self.ipr.route("del", **route)
            res.append(
                "ip route del {dst} dev {dev} table {table}".format(
                    dst=route['dst'],
                    dev=dev,
                    table=route['table'],
                )
            )
        return "\n".join(res)

    def get_route_entries(self, dests=None, table=None, dev=None):
        """
        Query for routes. Returns a dict suitable for applying (with **) to
        ipr's route del function.

        Unlike passing ipr.get_routes a dst argument, this only returns
        specific routes which actually exist (as opposed to synthesizing a
        route possibly from a larger match, and returning it with the wrong
        table, as get_routes does).
        """
        current_routes = self.ipr.get_routes()
        if dev:
            oif = self.ipr.link_lookup(ifname=dev)[0]
        # flatten attrs list to dict (api allows duplicate keys - but we don't)
        [r.update(attrs=dict(r['attrs'])) for r in current_routes]
        # install dst key with cidr notation
        [
            r.update(dst=r['attrs']['RTA_DST'] + "/" + str(r['dst_len']))
            for r in current_routes
            if 'RTA_DST' in r['attrs']
        ]
        res = [
            dict(
                dst=r['dst'],
                table=r['attrs']['RTA_TABLE'],
                oif=r['attrs']['RTA_OIF'],
                scope=r['scope'],
            )
            for r in current_routes
            if 'dst' in r
            and (dests is None or r['dst'] in dests)
            and (table is None or r['attrs']['RTA_TABLE'] == table)
            and (dev is None or r['attrs']['RTA_OIF'] == oif)
        ]
        return res

    def remove_unknown(self, dryrun=False):
        """
        This is currently the code path where disabled and removed peers get
        their routes and wg peer configs removed. In the future, deferred
        actions from the event engine should remove the specific things that we
        know need to be removed, and then this method will actually only be
        used to remove rogue entries.
        """
        # IPv6 analysis: not ipv6 ready
        # Please enhance this function to support ipv6
        routing_table = self.organize.table
        res = []
        enabled_pks = [
            str(peer.descriptor.pk)
            for peer in self.organize.peers.limit(enabled=True).values()
        ]
        for peer in self.wgi.peers:
            if peer['public_key'] not in enabled_pks:
                if not dryrun:
                    res.append(self.remove_wg_peer(peer['public_key'], dryrun))
                    self.log.info(
                        "Removing unexpected peer pk: (%s)", peer['public_key']
                    )
                res.append(
                    "wg set {interface} peer {pk} remove".format(
                        interface=self.wg_name, pk=peer['public_key']
                    )
                )
        expected_routes = [
            str(dst)
            for peer in self.organize.peers.limit(enabled=True).values()
            for dst in peer.allowed_ips
        ]

        current_routes = self.ipr.get_routes()
        [r.update(attrs=dict(r['attrs'])) for r in current_routes]
        our_current_routes = [
            r
            for r in current_routes
            if r['attrs']['RTA_TABLE'] == self.organize.table
        ]
        for route in our_current_routes:
            dst = route['attrs']['RTA_DST'] + "/" + str(route['dst_len'])
            if dst not in expected_routes:
                scope = route['scope']
                if not dryrun:
                    self.log.info("Removing unexpected route: (%s)", dst)
                    self.ipr.route(
                        'del', table=routing_table, dst=str(dst), scope=scope
                    )
                if scope in SCOPES:
                    # this is strictly cosmetic
                    # the printed "ip route" command is runnable with the scope
                    # as an integer too
                    scope = SCOPES[scope]
                res.append(
                    "ip route del {dst} table {table} scope {scope}".format(
                        dst=dst, table=routing_table, scope=scope
                    )
                )
        if not self.organize.peers.limit(use_as_gateway=True, enabled=True):

            default_routes = [
                r
                for r in current_routes
                if r['attrs'].get('RTA_TABLE') == _LINUX_MAIN_ROUTING_TABLE
                and r['attrs'].get('RTA_DST') in ('0.0.0.0', '128.0.0.0')
                and r['attrs'].get('RTA_DST')
            ]
            for route in default_routes:
                dst = route['attrs']['RTA_DST'] + "/" + str(route['dst_len'])
                scope = route['scope']
                if not dryrun:
                    self.log.info("Removing unexpected route: (%s)", dst)
                    self.ipr.route(
                        'del',
                        table=_LINUX_MAIN_ROUTING_TABLE,
                        dst=str(dst),
                        scope=scope,
                    )
                if scope in SCOPES:
                    # this is strictly cosmetic
                    # the printed "ip route" command is runnable with the scope
                    # as an integer too
                    scope = SCOPES[scope]
                res.append(
                    "ip route del {dst} table {table} scope {scope}".format(
                        dst=dst, table=_LINUX_MAIN_ROUTING_TABLE, scope=scope
                    )
                )

        return res

    def sync_routes(self, dests, table, dryrun=False):
        """
        Takes a list of cidr notation dests and a routing table, and ensures
        those routes are configured there. Returns a string.
        """
        res = []
        self.log.debug("looking for routes for: %r", dests)

        oif_idx = self.ipr.link_lookup(ifname=self.wg_name)

        system_state = self.organize.state.system_state

        for dest in map(ip_network, dests):
            routes = self.ipr.route("show", dst=str(dest), table=table)
            if not routes:
                src = None
                for net in system_state.current_subnets:
                    # note: current_subnets is consulted to find a source
                    # address but NOT consulted regarding the destination.
                    # (for pinned peers, we want to add IPs from
                    # non-current subnets here; they only need to be in a
                    # current subnet the first time they're seen)
                    if dest.version == net.version and dest.subnet_of(net):
                        src = system_state.current_subnets[net][0]
                        # select the first local IP we have in the first
                        # subnet. (it would be more correct to use the
                        # longest-prefix-matching subnet... but not
                        # bothering with that for now.)
                        break
                        # further note: this current_subnets logic
                        # doesn't belong here at all; will refactor
                        # this into system state soon.
                res.append(
                    f"ip route add {dest} dev {self.wg_name} proto "
                    f"static scope link%s table {table}"
                    % (f" src {src}" if src else "")
                )
                if not dryrun:
                    self.log.info("[#] %s", str(res[-1]))
                    self.ipr.route(
                        "add",
                        dst=str(dest),
                        oif=oif_idx,
                        table=table,
                        scope='link',
                        prefsrc=str(src) if src else None,
                    )
            else:
                for route in routes:
                    self.log.debug("found existing route for %s", dest)

        return "\n".join(res)



from unittest.mock import MagicMock

import vula.discover


class TestWireGuardServiceListener:
    def test_remove_service_does_call_callback(self):
        callback = MagicMock()
        listener = vula.discover.WireGuardServiceListener(callback)
        listener.remove_service(MagicMock(), "test_type", "test_name")
        callback.assert_not_called()

    def test_add_service_calls_callback(self):
        callback = MagicMock()
        props = {
            b'addrs': b'192.168.2.1',
            b'pk': b'EqcQ5gYxzGtzg7B4xi83kLyfuSMp8Kv3cmAJMs12nDM=',
            b'c': b'T6htsKgwCp5MAXjPiWxtVkccg+K2CePsVa7uyUgxE2ouYxKXg2qNL+'
            + b'0ut3sSVTYjzFGZSCO6n80SRaR+BIeOCg==',
            b'hostname': b'myhost',
            b'port': b'5054',
            b'vk': b'EqcQ5gYxzGtzg7B4xi83kLyfuSMp8Kv3cmAJMs12nDM=',
            b'dt': b'34',
            b'vf': b'42',
            b'r': b'192.168.2.0/24',
            b'e': b'1',
        }
        zeroconf = MagicMock()
        zeroconf.get_service_info().properties = props
        listener = vula.discover.WireGuardServiceListener(callback)

        listener.add_service(zeroconf, "test_type", "test_name")

        callback.assert_called_once()

    def test_add_service_no_service_info(self):
        callback = MagicMock()
        zeroconf = MagicMock()
        zeroconf.get_service_info.return_value = None
        listener = vula.discover.WireGuardServiceListener(callback)

        listener.add_service(zeroconf, "test_type", "test_name")

        callback.assert_not_called()

    def test_add_service_invalid_descriptor(self):
        callback = MagicMock()
        props = {
            b'addrs': b'192.168.2.1',
            b'hostname': b'myhost',
            b'port': b'5054',
            b'dt': b'34',
            b'r': b'192.168.2.0/24',
            b'e': b'1',
        }
        zeroconf = MagicMock()
        zeroconf.get_service_info().properties = props
        listener = vula.discover.WireGuardServiceListener(callback)

        listener.add_service(zeroconf, "test_type", "test_name")

        callback.assert_not_called()

    def test_update_service_calls_add_service(self):
        listener = vula.discover.WireGuardServiceListener(MagicMock())
        listener.add_service = MagicMock()
        m = MagicMock()

        listener.update_service(m, "bar", "foo")

        listener.add_service.assert_called_once_with(m, "bar", "foo")


def alive_mock(x):
    m = MagicMock()
    m.is_alive.return_value = x
    return m


class TestDiscover:
    def test_callback_calls_all(self):
        a = MagicMock()
        b = MagicMock()
        c = MagicMock()

        discover = vula.discover.Discover()
        discover.callbacks = [a, b, c]

        discover.callback("foo")

        a.assert_called_once_with("foo")
        b.assert_called_once_with("foo")
        c.assert_called_once_with("foo")

    def test_is_not_alive(self):
        discover = vula.discover.Discover()
        discover.browsers = {
            '192.168.1.1': alive_mock(False),
            '192.168.1.2': alive_mock(False),
            '192.168.1.3': alive_mock(False),
        }

        alive = discover.is_alive()
        print(alive)
        assert alive is False

    def test_is_alive(self):
        discover = vula.discover.Discover()
        discover.browsers = {
            '192.168.1.1': alive_mock(False),
            '192.168.1.2': alive_mock(False),
            '192.168.1.3': alive_mock(True),
            '192.168.1.4': alive_mock(False),
        }

        alive = discover.is_alive()
        assert alive is True

    def test_shutdown_removes_browsers(self):
        discover = vula.discover.Discover()
        discover.browsers = {
            '192.168.1.1': MagicMock(),
            '192.168.1.2': MagicMock(),
            '192.168.1.3': MagicMock(),
        }

        discover.shutdown()

        assert discover.browsers == {}

    def test_shutdown_cancels_browsers(self):
        browser1 = MagicMock()
        browser2 = MagicMock()
        browser3 = MagicMock()
        discover = vula.discover.Discover()
        discover.browsers = {
            '192.168.1.1': browser1,
            '192.168.1.2': browser2,
            '192.168.1.3': browser3,
        }

        discover.shutdown()

        browser1.cancel.assert_called_once()
        browser2.cancel.assert_called_once()
        browser3.cancel.assert_called_once()



import hashlib
import os
import sys

import click

try:
    import ggwave
except ImportError:
    ggwave = None

try:
    import pyaudio
except ImportError:
    pyaudio = None

_GGWAVE_VOLUME: int = (
    100  # Full ggwave volume, use the os controls for adjustments
)


class VerifyAudio:
    def __init__(self, verbose):
        self.verbose = verbose
        self.null_fds = None
        self.store_fds = None

        if not self.verbose:
            self.__mute()

        try:
            self.py_audio = pyaudio.PyAudio()
        except AttributeError:
            click.echo("pyaudio error")
            sys.exit(4)

        if not self.verbose:
            self.__unmute()

        try:
            self.ggwave_instance = ggwave.init()
        except AttributeError:
            click.echo("ggwave error")
            sys.exit(5)

    def send_verification_key(self, vk):
        if self.verbose:
            click.echo(
                "Speaking hash: "
                + hashlib.sha256(vk.encode('utf-8')).hexdigest()
            )

        try:
            waveform = ggwave.encode(
                hashlib.sha256(vk.encode('utf-8')).hexdigest(),
                protocolId=1,
                volume=_GGWAVE_VOLUME,
            )
        except AttributeError:
            click.echo("ggwave error")
            sys.exit(5)

        stream = self.py_audio.open(
            format=pyaudio.paFloat32,
            channels=1,
            rate=48000,
            output=True,
            frames_per_buffer=4096,
        )
        try:
            stream.write(waveform, len(waveform) // 4)
            stream.stop_stream()
        except SystemError as e:
            click.echo(f"pyaudio error: {e}")
            sys.exit(5)
        finally:
            stream.close()
            self.py_audio.terminate()

    def receive_verification_key(self):
        stream = self.py_audio.open(
            format=pyaudio.paFloat32,
            channels=1,
            rate=48000,
            input=True,
            frames_per_buffer=1024,
        )
        try:
            count = 0
            while True:
                count += 1
                data = stream.read(1024, exception_on_overflow=False)
                click.echo(f"receiving audio data...{count}\r", nl=False)
                if not self.verbose:
                    self.__mute()
                res = ggwave.decode(self.ggwave_instance, data)
                if not self.verbose:
                    self.__unmute()
                if res is not None:
                    try:
                        return res.decode("utf-8")
                    except ValueError:
                        return None

        except KeyboardInterrupt:
            return None
        finally:
            ggwave.free(self.ggwave_instance)
            stream.stop_stream()
            stream.close()
            self.py_audio.terminate()

        return None

    def __mute(self):
        # Open a pair of null files
        self.null_fds = [os.open(os.devnull, os.O_RDWR) for x in range(2)]

        # Save the actual stdout (1) and stderr (2) file descriptors.
        self.store_fds = [os.dup(1), os.dup(2)]

        # Redirect descriptors
        os.dup2(self.null_fds[0], 1)
        os.dup2(self.null_fds[1], 2)

    def __unmute(self):
        # Reverse redirect
        os.dup2(self.store_fds[0], 1)
        os.dup2(self.store_fds[1], 2)

        # Make sure all fds are closed
        for fd in self.null_fds + self.store_fds:
            os.close(fd)



"""
This module provides the interface for configuring WireGuard interfaces.

It currently requires PyRoute2 and only works on GNU/Linux; in the future it
should contain implementations for other platforms too.

A goal of the Feb12 2021 refactor is that sys_pyroute2 will be the only user of
this module.
"""

from __future__ import annotations

import time
from base64 import b64decode, b64encode  # noqa: F401
from datetime import timedelta
from ipaddress import ip_address, ip_network
from logging import Logger, getLogger
from typing import Tuple

import click
from pyroute2 import IPRoute
from pyroute2 import WireGuard as PyRoute2WireGuard
from pyroute2.netlink import nla as netlink_atom
from schema import And, Optional, Or, Schema, Use

from .common import (
    DualUse,
    attrdict,
    b64_bytes,
    comma_separated_Nets,
    format_byte_stats,
    jsonrepr_hl,
    schemattrdict,
    serializable,
    yamlrepr_hl,
)


def _wg_interface_list():
    """
    This returns a list of the current wireguard interfaces' names.

    There must be a better way to do this!
    >>> type(_wg_interface_list())
    <class 'list'>
    """
    interfaces = []
    links = IPRoute().get_links()
    for link in links:
        linkinfo = link.get_attr('IFLA_LINKINFO')
        if linkinfo is not None:
            if linkinfo.get_attr('IFLA_INFO_KIND') == "wireguard":
                interfaces.append(link.get_attr('IFLA_IFNAME'))
    return interfaces


class PeerConfig(schemattrdict, serializable):

    schema = Schema(
        {
            Optional('unspec'): object,
            Optional('remove'): bool,
            Optional('public_key'): And(b64_bytes.with_len(32), Use(str)),
            Optional('preshared_key'): And(b64_bytes.with_len(32), Use(str)),
            Optional('endpoint_addr'): And(ip_address, Use(str)),
            Optional('endpoint_port'): Use(int),
            Optional('persistent_keepalive'): Use(int),
            Optional('allowed_ips'): Or(
                [And(ip_network, Use(str))],
                And(Use(comma_separated_Nets), Use(list)),
            ),
            Optional('stats'): dict(
                rx_bytes=int, tx_bytes=int, latest_handshake=int
            ),
            Optional('latest_handshake'): Use(int),
        }
    )

    default = dict(remove=False)

    @classmethod
    def from_netlink(cls, peer):
        """
        This converts approximately from what pyroute2 produces to what
        pyroute2 consumes

        (plus the extra key 'stats' with two keys)

        >>> p = PeerConfig.from_netlink({'attrs': dict(
        ...     public_key=b64encode(b'A'*32),
        ...     preshared_key=b64encode(b'A'*32),
        ...     persistent_keepalive_interval=666,
        ...     rx_bytes=2,
        ...     tx_bytes=3,
        ...     protocol_version=99,
        ...     endpoint={'addr': '192.168.0.0', 'port': 1000},
        ...     last_handshake_time={'tv_sec':567})})
        >>> type(p)
        <class 'vula.wg.PeerConfig'>
        >>> p['persistent_keepalive']
        666
        >>> p['endpoint_addr']
        '192.168.0.0'
        >>> p['endpoint_port']
        1000
        >>> p['stats']['latest_handshake']
        567
        >>> p['protocol_version'] # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        KeyError:
        >>> p = PeerConfig.from_netlink({'attrs': dict(
        ...     public_key=b64encode(b'A'*32),
        ...     preshared_key=b64encode(b'A'*32),
        ...     persistent_keepalive_interval=666,
        ...     rx_bytes=2,
        ...     tx_bytes=3,
        ...     protocol_version=99,
        ...     endpoint={'addr': 'FE80::FFFF:FFFF:FFFF:FFFE', 'port': 1000},
        ...     last_handshake_time={'tv_sec':567})})
        >>> type(p)
        <class 'vula.wg.PeerConfig'>
        >>> p['persistent_keepalive']
        666
        >>> p['endpoint_addr']
        'FE80::FFFF:FFFF:FFFF:FFFE'
        >>> p['endpoint_port']
        1000
        >>> p['stats']['latest_handshake']
        567
        >>> p['protocol_version'] # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        KeyError:
        """
        res = {
            k.replace('WGPEER_A_', '').lower(): dict(v)
            if isinstance(v, netlink_atom)
            else v
            for k, v in dict(peer['attrs']).items()
        }
        res['allowed_ips'] = [
            ip_network(
                (
                    bytes.fromhex(
                        net['WGALLOWEDIP_A_IPADDR'].replace(':', '')
                    ),
                    net['WGALLOWEDIP_A_CIDR_MASK'],
                ),
            )
            for net in (
                dict(atom['attrs']) for atom in res.pop('allowedips', ())
            )
        ]
        res['public_key'] = res['public_key'].decode()
        res['preshared_key'] = res['preshared_key'].decode()
        if b64decode(res['preshared_key']) == bytes([0]) * 32:
            del res['preshared_key']

        if 'endpoint' in res:
            res['endpoint_addr'] = res['endpoint']['addr']
            res['endpoint_port'] = res['endpoint']['port']
            res.pop('endpoint')
        res['persistent_keepalive'] = res.pop('persistent_keepalive_interval')
        res['stats'] = dict(
            rx_bytes=res['rx_bytes'],
            tx_bytes=res['tx_bytes'],
            latest_handshake=res.pop('last_handshake_time')['tv_sec'],
        )
        res.pop('rx_bytes')
        res.pop('tx_bytes')
        res.pop('protocol_version')
        return cls(res)

    @property
    def wg_show(self: PeerConfig):
        return "\n  ".join(
            click.style(label, bold=True) + ': ' + value
            for label, value in {
                click.style('peer', fg="yellow"): click.style(
                    self['public_key'], fg="yellow"
                ),
                'preshared key': (self.get('preshared_key') and '(hidden)'),
                'endpoint': (
                    'endpoint_addr' in self
                    and 'endpoint_port' in self
                    and "{endpoint_addr}:{endpoint_port}".format(**self)
                ),
                'allowed ips': (", ".join(self['allowed_ips']) or "(none)"),
                'latest handshake': (
                    self['stats']['latest_handshake']
                    and str(
                        timedelta(
                            seconds=int(
                                time.time() - self['stats']['latest_handshake']
                            )
                        ),
                    )
                    + ' ago'
                ),
                'transfer': (
                    sum(self['stats'].values())
                    and "{rx_bytes} received, {tx_bytes} sent".format(
                        **format_byte_stats(self['stats'])
                    )
                ),
                'persistent keepalive': (
                    self.get('persistent_keepalive')
                    and "every %s seconds" % (self['persistent_keepalive'],)
                ),
            }.items()
            if value not in (None, False)
        )

    @property
    def wg_showconf(self: PeerConfig):
        return "[Peer]\n" + "\n".join(
            label + ' = ' + value
            for label, value in {
                'PublicKey': self['public_key'],
                'PresharedKey': self.get('preshared_key'),
                'AllowedIPs': ", ".join(self['allowed_ips']),
                'Endpoint': (
                    'endpoint_addr' in self
                    and 'endpoint_port' in self
                    and "{endpoint_addr}:{endpoint_port}".format(**self)
                ),
                'PersistentKeepalive': (
                    self.get('persistent_keepalive')
                    and str(self['persistent_keepalive'])
                ),
            }.items()
            if value not in (None, False, '')
        )


@DualUse.object()
@click.argument('name', type=str)
class Interface(attrdict, yamlrepr_hl):

    """
    This is a wrapper for pyroute2's WireGuard interface.

    Give it an interface name, and it will give you a dict with the data in the
    same shape as the pyroute2 structure but with less-irritating names.

    But wait, there's more... it also has a "peers" attribute which gives you
    the peers in the same shape as pyroute2's Wireguard module expects them to
    be passed to the "set" method (note: round trips not yet tested).

    Because this is a DualUse.object, you can see the data on the commandline
    with commands like these:

    sudo vula wg Interface vula query

    sudo vula wg Interface vula peers

    sudo vula wg Interface vula _is_up

    etc.
    """

    def __init__(self, name, ipr=None):
        """
        >>> iface = Interface("test interface")
        >>> iface.name
        'test interface'
        >>> iface._ipr # doctest: +ELLIPSIS
        <pyroute2.iproute.linux.IPRoute object at 0x...>
        >>> len(Interface(42))
        0
        """
        self.log: Logger = getLogger()
        self.name = name
        self._wg = PyRoute2WireGuard()
        if ipr is None:
            ipr = IPRoute()
        self._if_index = None
        self._ipr = ipr
        self.query()

    @property
    def _get_link(self):
        """
        >>> custom_ipr = IPRoute()
        >>> link_name = custom_ipr.get_links(0)[0].get_attr('IFLA_IFNAME')
        >>> int = Interface(link_name, custom_ipr)
        >>> link_name == int._get_link.get_attr('IFLA_IFNAME')
        True

        >>> int = Interface("random interface name")
        >>> int._get_link
        {}
        """
        res = self._ipr.get_links(ifname=self.name)
        if res:
            return res[0]
        else:
            return {}

    @property
    def _exists(self):
        return bool(self._ipr.link_lookup(ifname=self.name))

    @property
    def _is_up(self):
        return bool(self._get_link.get('state') == 'up')

    @DualUse.method()
    @click.option('-n', '--dryrun', is_flag=True)
    @click.argument('private_key', type=str)
    @click.argument('listen_port', type=int)
    @click.argument('fwmark', type=int)
    def sync_interface(self, private_key, listen_port, fwmark, dryrun):
        """
        Creates, brings up, and configures an interface.

        Returns a list of the necessary steps.
        """
        res = []

        private_key = private_key.encode()

        if not self._exists:
            if not dryrun:
                self._ipr.link('add', ifname=self.name, kind="wireguard")
            res.append('# create interface')
            res.append('ip link add %s type wireguard' % (self.name,))

        if not self._is_up:
            if not dryrun:
                if_index = self._ipr.link_lookup(ifname=self.name)[0]
                self._ipr.link('set', index=if_index, state='up')
            res.append('# bring up interface')
            res.append('ip link set up %s' % (self.name,))

        self.query()

        data = dict(
            private_key=private_key, listen_port=listen_port, fwmark=fwmark
        )
        todo = {}
        for k, v in data.items():
            if self.get(k) != v:
                todo[k] = v
        if todo:
            if not dryrun:
                self.set(**dict(todo))
            if 'private_key' in todo:
                todo['private_key'] = '<redacted private key>'
            res.append('# configure interface')
            res.append("WireGuard.set(%r, **%r)" % (self.name, todo))
        return res

    @DualUse.method()
    def query(self):
        """
        This calls "info" for the interface via pyroute2, and (re-)populates
        our dictionary (we're a dict subclass, recall). It returns self.
        """
        self.clear()
        self.log.debug("Fetching interface info for %s", self.name)
        try:
            res: Tuple = self._wg.info(self.name)
        except Exception as ex:
            self.log.debug("Failed to query interface %r: %r", self.name, ex)
            return self

        res = {
            k.replace('WGDEVICE_A_', '').lower(): dict(v)
            if isinstance(v, netlink_atom)
            else v
            for k, v in dict(res[0]['attrs']).items()
        }
        res['peers'] = list(map(PeerConfig.from_netlink, res.get('peers', ())))
        self.clear()
        self.update(res)
        return self

    def set(self, **kwargs):
        self.log.debug("Calling WireGuard.set(%r, **%r)", self.name, kwargs)
        res = self._wg.set(self.name, **kwargs)
        self.log.debug("WireGuard.set(%r, **%r) -> %r", self.name, kwargs, res)
        return res

    def apply_peerconfig(self, new: attrdict, dryrun=False):
        """
        This sets only the keys that have changed, and returns a list of the
        new keys that needed to be set. Due to a bug in PyRoute2 and/or Linux,
        it is necessary to always set the allowed_ips if anything is set, so
        this does that.
        """
        self.query()
        cur = self._peers_by_pubkey.get(new["public_key"])
        res = []
        if cur:

            if new.get('remove'):
                res.append(
                    '# removing wireguard peer %s' % (new['public_key'],)
                )

            elif 'allowed_ips' not in new:
                # pyroute2/wg bug workaround
                new['allowed_ips'] = cur['allowed_ips']

            for key in list(new):
                if cur.get(key) == new[key]:
                    if key in ("allowed_ips", 'public_key'):
                        # allowed_ips is here for pyroute2/wg bug workaround
                        continue
                    self.log.debug(
                        "not resetting %r=%r as it is unchanged",
                        key,
                        cur.get(key),
                    )
                    del new[key]
                elif key != 'remove':
                    res.append(
                        "# {key} is {cur}; should be {new}".format(
                            key=key, cur=cur.get(key), new=new[key]
                        )
                    )
            # workaround for pyroute2 irritatingly handling endpoint addr
            # and port separately (while they actually need to be set
            # together)
            if 'endpoint_addr' in new and 'endpoint_port' not in new:
                new['endpoint_port'] = cur['endpoint_port']
            elif 'endpoint_port' in new and 'endpoint_addr' not in new:
                new['endpoint_addr'] = cur['endpoint_addr']
        else:
            if new.get('remove'):
                return "# can't remove non-existent wireguard peer %s" % (
                    new['public_key'],
                )

        if (
            cur
            and len(new) == 2
            and new.get('allowed_ips') == cur['allowed_ips']
        ):
            # pyroute2/wg bug workaround
            self.log.debug("apply_peerconfig: no wg update necessary")

        else:
            if cur:
                res.append(
                    '# reconfigure wireguard peer %s' % (new['public_key'],)
                )
            else:
                res.append(
                    '# configure new wireguard peer %s' % (new['public_key'],)
                )

            res.append(
                "vula wg set {interface} peer {pk} "
                "{remove}{endpoint}{args}{allowed_ips}".format(
                    remove="remove " if new.get('remove') else "",
                    endpoint="endpoint %s:%s "
                    % (new['endpoint_addr'], new['endpoint_port'])
                    if (new.get('endpoint_addr') and new.get('endpoint_port'))
                    else '',
                    args="".join(
                        "%s %s " % (k, v)
                        for k, v in new.items()
                        if k in ('persistent_keepalive', 'preshared_key')
                    ),
                    allowed_ips='allowed-ips %s '
                    % ",".join(ip for ip in new.get('allowed_ips', ()))
                    if 'allowed_ips' in new
                    else "",
                    interface=self.name,
                    pk=new['public_key'],
                )
            )

            for line in res:
                self.log.info("[#] %s", line)

            if not dryrun:
                self.set(peer=new)

        res = "\n".join(filter(None, res))
        return res

    @property
    def peers(self):
        """
        Returns list of peer structures which should be identical to those
        passed to the pyroute2 Wireguard set method, except with an extra
        "stats" key.

        And also except for that it won't do IPv6 correctly.

        This would be a good function to write tests for (and perhaps send
        upstream to pyroute2...)
        """
        return self.get('peers', [])

    @property
    def _peers_by_pubkey(self):
        return {peer['public_key']: peer for peer in self.peers}

    @property
    def wg_show(self) -> str:
        """
        This returns output similar to the "wg show" command.
        """
        peers = list(self.peers)
        return (
            "\n  ".join(
                click.style(label, bold=True) + ': ' + str(value)
                for label, value in {
                    click.style(
                        'interface', fg="green", bold=True
                    ): click.style(self.name, fg="green"),
                    'public key': (
                        'public_key' in self and self['public_key'].decode()
                    ),
                    'private key': ('public_key' in self and '(hidden)'),
                    'listening port': self.get('listen_port'),
                    'fwmark': (
                        self.get('fwmark') and "0x%x" % (self['fwmark'],)
                    ),
                }.items()
                if value not in (None, False)
            )
            + ("\n\n" if peers else '')
            + "\n\n".join(peer.wg_show for peer in peers)
        )

    @property
    def wg_showconf(self) -> str:
        """
        Shows the current configuration of a given WireGuard interface, for use
        with `setconf'.
        """
        peers = list(self.peers)
        return "[Interface]\n" + (
            "\n".join(
                "%s = %s" % (label, value)
                for label, value in {
                    'ListenPort': self.get('listen_port'),
                    'FwMark': (
                        self.get('fwmark') and "0x%x" % (self['fwmark'],)
                    ),
                    'PrivateKey': self.get('private_key', b'').decode(),
                }.items()
                if value not in (None, False, '')
            )
            + ("\n\n" if peers else '')
            + "\n\n".join(peer.wg_showconf for peer in peers)
        )


@DualUse.object(
    invoke_without_command=True,
    short_help="set and retrieve configuration of WireGuard interfaces",
)
@click.pass_context
class wg(object):
    """
    The subcommands here exist mostly to aid in the development of the
    wg.Interface class, and aren't currently intended for normal use.
    """

    def __init__(self, ctx, *a, **kw):
        if ctx.invoked_subcommand is None:
            click.echo(self.show())

    @DualUse.method(
        short_help="Shows the current configuration and device information"
    )
    @click.option(
        '-f',
        '--format',
        'fmt',
        type=click.Choice(["wireguard", "json", "yaml"]),
        default='wireguard',
        show_default=True,
    )
    @click.argument('interfaces', type=str, nargs=-1)
    def show(self, fmt="wireguard", interfaces=()):
        """
        Produces output very similar to the "wg show" command.

        (The transfer counters, last handshake time, and keepalive interval are
        all formatted differently.)
        """
        if len(interfaces) == 0:
            interfaces = _wg_interface_list()

        interfaces = [Interface(name).query() for name in interfaces]

        if fmt == 'wireguard':
            return "\n\n".join(iface.wg_show for iface in interfaces)

        elif fmt == 'yaml':
            return str(
                yamlrepr_hl({iface.name: iface for iface in interfaces})
            )

        elif fmt == 'json':
            return str(
                jsonrepr_hl({iface.name: iface for iface in interfaces})
            )

    @DualUse.method()
    @click.argument('interface', type=str)
    def showconf(self, interface):
        """
        Shows the current configuration of a given WireGuard interface, for use
        with `setconf'.
        """
        return Interface(interface).query().wg_showconf

    @DualUse.method(
        short_help="Change the current configuration, add peers, remove "
        "peers, or change peers."
    )
    @click.option('--private-key', type=str)
    @click.option('--listen-port', type=int)
    @click.option('--fwmark', type=int)
    @click.argument('interface', type=str)
    @click.argument('args', type=str, nargs=-1)
    @click.pass_context
    def set(ctx, self, interface, args=(), **kwargs):
        """
        Change the current configuration, add peers, remove peers, or change
        peers.

        Usage: wg set <interface> [listen-port <port>] [fwmark <mark>]
        [private-key <base64 private key>] [peer <base64 public key> [remove]
        [preshared-key <base64 preshared key>] [ endpoint <ip>:<port>]
        [persistent-keepalive <interval seconds>] [allowed-ips
        <ip1>/<cidr1>[,<ip2>/<cidr2>]...] ]...

        This is intended to behave very similarly to the normal wg tool, except
        it takes private keys on the commandline (and does not yet support
        reading them from files, as wg does). Yes, this is not a great idea,
        but it makes testing easier.
        """
        # IPv6 analysis: not ipv6 ready.
        # Please enhance this function to support ipv6
        dev = Interface(interface)
        kwargs = {k: v for k, v in kwargs.items() if v not in (None, ())}
        current = {}
        kwargss = []
        args = list(sum(kwargs.items(), ())) + list(args)
        allowed = {"private_key", "listen_port", "fwmark"}
        try:
            while args:
                key = args.pop(0)
                key = key.replace('-', '_')
                if key in allowed:
                    allowed.remove(key)
                    if current.get(key) is not None:
                        raise Exception("duplicate argument: %r" % (key,))
                    current[key] = args.pop(0)
                elif key == "peer":
                    if current:
                        kwargss.append(current)
                    pk = args.pop(0)
                    current = dict(public_key=pk)
                    allowed = {
                        "preshared_key",
                        "endpoint",
                        "persistent_keepalive",
                        "allowed_ips",
                    }
                elif key == "remove":
                    if not kwargss:
                        raise Exception("'remove' requires 'peer' first")
                    current[key] = True
                else:
                    raise Exception("bad key: %r" % (key,))
            kwargss.append(current)
        except Exception as ex:
            raise Exception("Failed to parse arguments: %r" % (ex,))
        res = []
        if 'public_key' not in kwargss[0]:
            res.append(dev.set(**kwargss.pop(0)))
        for peer in kwargss:
            if 'endpoint' in peer:
                peer['endpoint_addr'], peer['endpoint_port'] = peer[
                    'endpoint'
                ].split(':')
                del peer['endpoint']
            peer = PeerConfig(peer)._dict()
            res.append(dev.apply_peerconfig(peer))
        return "\n".join(map(str, res))

    @DualUse.method()
    @click.argument('interface', type=str)
    @click.argument('public_key', type=str)
    @click.option('--preshared-key', type=str)
    @click.option('--endpoint-addr', type=lambda x: str(ip_address(x)))
    @click.option('--endpoint-port', type=int)
    @click.option('--persistent-keepalive', type=int)
    @click.option(
        '--allowed-ip',
        'allowed_ips',
        type=lambda x: str(ip_network(x)),
        multiple=True,
    )
    @click.option('--remove', is_flag=True, help="Remove this peer")
    def set_peer(self, interface, **kwargs):
        """
        This allows setting peer information.
        """
        kwargs = {k: v for k, v in kwargs.items() if v not in (None, ())}
        if 'allowed_ips' in kwargs:
            kwargs['allowed_ips'] = list(kwargs['allowed_ips'])
        wg = Interface(interface)
        return wg.apply_peerconfig(attrdict(kwargs))


@click.group()
def link():
    """
    Link commands
    """


@link.command(name='del', short_help="Delete an interface")
@click.argument('name', type=str)
def del_(name):
    """
    Delete an interface

    Note: this is not wireguard-specific.
    """
    IPRoute().link('del', ifname=name)


@link.command()
@click.argument('name', type=str)
def add(name):
    """
    Add a new wireguard interface
    """
    IPRoute().link('add', ifname=name, kind="wireguard")


wg.cli.add_command(link, name="link")
wg.cli.add_command(Interface.cli, name="Interface")

main = wg.cli
